#
# QosCosGrid Computing
# Copyright (C) 2008-2011 Poznan Supercomputing and Networking Center
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>
#

from lsf import *
from datetime import datetime
import time
from InfoStorage import HostInfo, QueueInfo, JobInfo
from ClusterMgmt import ClusterMgmt
from xml.dom.minidom import Document

class LsfImpl(ClusterMgmt):
	def __init__(self):
		self.__lsf_cluster = LsfCluster()
	def getHostInfo(self, Hosts=None):
		"""Gets information about hosts in cluster
		Hosts, string list, contains names of hosts of interest
			if None, all hosts are considered
		Returns array of HostInfo structures
		"""	
		lsfCluster = self.__lsf_cluster	
		hosts = lsfCluster.HostInfo(0, Hosts)		
		ret = []
		for host in hosts:        
			hostInfo = HostInfo()
			name = host["hostName"]        
			extHostInfo = lsfCluster.HostInfoLsb([name])[0]
			hostLoadInfo = lsfCluster.LoadInfo(0, [name])[0]
			hostResources = lsfCluster.SharedResource(0, name)
			hostQueues = lsfCluster.QueueInfoEnt(0, None, name)
			hostJobs = lsfCluster.JobInfoEnt(0, 0, None, None, None, name)
			
			hostInfo.HostName = name
			if "cpuarch" in hostResources:
				hostInfo.CpuArch = hostResources["cpuarch"]
			else:
				hostInfo.CpuArch = "Other"
			if "osname" in hostResources:
				hostInfo.OSName = hostResources["osname"]
			else:
				hostInfo.OSName = "Unknown"
			if "cpuspeed" in hostLoadInfo:
				hostInfo.CpuSpeed = hostLoadInfo["cpuspeed"]
			else:
				hostInfo.CpuSpeed = 0.0
						
			hostInfo.CpuCount = host["maxCpus"]
			hostInfo.CpuUtilization = hostLoadInfo["ut"]
			hostInfo.QueueLength15S = hostLoadInfo["r15s"]
			hostInfo.QueueLength1M = hostLoadInfo["r1m"]
			hostInfo.QueueLength15M = hostLoadInfo["r15m"]
			hostInfo.PagingActivity = hostLoadInfo["pg"]
			hostInfo.FreeMemory = hostLoadInfo["mem"]
			hostInfo.FreeSwap = hostLoadInfo["swp"]
			hostInfo.FreeFilesystem = hostLoadInfo["tmp"]
			hostInfo.JobNum = extHostInfo["numJobs"]
			hostInfo.JobMax = extHostInfo["maxJobs"]
			Jobs = []
			for job in hostJobs:
				Jobs.append(job["jobId"])
				Queues = []
			for queue in hostQueues:
				Queues.append(queue["queue"])
			hostInfo.Jobs = Jobs
			hostInfo.Queues = Queues
			hostInfo.AdditionalResources["lsf"] = host["resources"]
			ret.append(hostInfo)
		return ret
	
	def getQueueInfo(self, Queues=None):
		"""Gets information about queues in cluster
		Queues, string list, contains names of queues of interest
			if None, all queues are considered
		Returns array of QueueInfo structures
		"""
		lsfCluster = self.__lsf_cluster
		queues = lsfCluster.QueueInfoEnt(0, Queues)
		ret = []
		for queue in queues:
			name = queue["queue"]
			queueJobs = lsfCluster.JobInfoEnt(0, 0, None, None, name)
			queueInfo = QueueInfo()
			queueInfo.QueueName = name
			queueInfo.QueuePriority = queue["priority"]			
			queueInfo.QueueEnabled = ((queue["qStatus"] & 4) == 4)
			queueInfo.PendingJobs = queue["numPEND"]
			queueInfo.RunningJobs = queue["numRUN"]
			queueInfo.MaxJobs = queue["maxJobs"]
			Jobs = []
			for job in queueJobs:
				Jobs.append(job["jobId"])
			queueInfo.JobIDs = Jobs
			queueInfo.AdditionalInfo["description"] = queue["description"]
			ret.append(queueInfo)
		return ret

	def getJobInfo(self, JobID=0):
		"""Gets information about queues in cluster
		JobID, long integer, specifies job of interest
			all jobs are considered if JobID=0
		Returns array of JobInfo structures
		"""
		lsfCluster = self.__lsf_cluster
		jobs = lsfCluster.JobInfoEnt(0, JobID)
		ret = []
		for job in jobs:
			jobInfo = JobInfo()
			jobInfo.JobID = int(job["jobId"])
			jobInfo.JobName = job["jobName"]
			jobInfo.OwnerName = job["user"]
			status = job["status"]
			if (status & 1 == 1):
				jobInfo.JobStatus = "Pending"
			elif (status & 2 == 2):
				jobInfo.JobStatus = "Suspended"
			elif (status & 4 == 4):
				jobInfo.JobStatus = "Running"
			elif (status & 8 == 8):
				jobInfo.JobStatus = "Suspended"
			elif (status & 16 == 16):
				jobInfo.JobStatus = "Suspended"
			elif (status & 32 == 32):
				jobInfo.JobStatus = "Exited"
			elif (status & 64 == 64):
				jobInfo.JobStatus = "Done"
			else:
				jobInfo.JobStatus = "Unknown"			
			jobInfo.QueueName = job["queue"]
			jobInfo.SubmissionTime = datetime.fromtimestamp(job["submitTime"])
			time = job["startTime"]
			if time == 0:
				jobInfo.StartTime = None
			else:
				jobInfo.StartTime = datetime.fromtimestamp(time)
			time = job["endTime"]
			if time == 0:
				jobInfo.EndTime = None
			else:
				jobInfo.EndTime = datetime.fromtimestamp(time)
			jobInfo.ExecutionHosts = job["exHosts"]
			if job["mem"] < 0:
				jobInfo.TotalMem = None
			else:
				jobInfo.TotalMem = job["mem"]		
			if job["swap"] < 0:
				jobInfo.TotalSwap = None
			else:
				jobInfo.TotalSwap = job["swap"]
			if job["utime"] < 0:
				jobInfo.UserTime = None
			else:
				jobInfo.UserTime = job["utime"]
			if job["stime"] < 0:
				jobInfo.SystemTime = None
			else:
				jobInfo.SystemTime = job["stime"]
			
			ret.append(jobInfo)
		return ret

	def besXml(self):
		"""Gets information about hosts in XML format (bes-factory namespace)."""
		doc = Document()

		cluster = doc.createElementNS(ClusterMgmt.ClusterMgmtNS, ClusterMgmt.ClusterMgmtNSPrefix + ":Cluster")
		cluster.setAttribute("type", "SGE")
		doc.appendChild(cluster)		

		hosts = self.getHostInfo()
		for host in hosts:
			containedResource = doc.createElement("ContainedResource")
			cluster.appendChild(containedResource)

			resourceName = doc.createElementNS(ClusterMgmt.BesFactoryNS, ClusterMgmt.BesFactoryNSPrefix + ":ResourceName")
			containedResource.appendChild(resourceName)
			text = doc.createTextNode(str(host.HostName))
			resourceName.appendChild(text)

			operatingSystem = doc.createElementNS(ClusterMgmt.BesFactoryNS, ClusterMgmt.BesFactoryNSPrefix + ":OperatingSystem")
			containedResource.appendChild(operatingSystem)
			operatingSystemType = doc.createElementNS(ClusterMgmt.JsdlNS, ClusterMgmt.JsdlNSPrefix + ":OperatingSystemType")
			operatingSystem.appendChild(operatingSystemType)
			operatingSystemName = doc.createElementNS(ClusterMgmt.JsdlNS, ClusterMgmt.JsdlNSPrefix + ":OperatingSystemName")
			operatingSystemType.appendChild(operatingSystemName)
			text = doc.createTextNode(str(host.OSName))
			operatingSystemName.appendChild(text)

			cpuArchitecture = doc.createElementNS(ClusterMgmt.BesFactoryNS, ClusterMgmt.BesFactoryNSPrefix + ":CPUArchitecture")
			containedResource.appendChild(cpuArchitecture)
			cpuArchitectureName = doc.createElementNS(ClusterMgmt.JsdlNS, ClusterMgmt.JsdlNSPrefix + ":CPUArchitectureName")
			cpuArchitecture.appendChild(cpuArchitectureName)
			text = doc.createTextNode(str(host.CpuArch))
			cpuArchitectureName.appendChild(text)

			cpuCount = doc.createElementNS(ClusterMgmt.BesFactoryNS, ClusterMgmt.BesFactoryNSPrefix + ":CPUCount")
			containedResource.appendChild(cpuCount)
			text = doc.createTextNode(str(int(host.CpuCount)))
			cpuCount.appendChild(text)

			cpuSpeed = doc.createElementNS(ClusterMgmt.BesFactoryNS, ClusterMgmt.BesFactoryNSPrefix + ":CPUSpeed")
			containedResource.appendChild(cpuSpeed)
			text = doc.createTextNode(str(int(host.CpuSpeed * 1024 * 1024)))
			cpuSpeed.appendChild(text)

		return doc
		
	def getAllInfoXml(self, detailedHostList=False, detailedJobList=False):
		"""Gets information about this cluster in XML format.
		See schema for format details.
		"""
		doc = Document()
		
		cluster = doc.createElementNS(ClusterMgmt.ClusterMgmtNS, ClusterMgmt.ClusterMgmtNSPrefix + ":Cluster")
		cluster.setAttribute("type", "LSF")
		doc.appendChild(cluster)
		
		xmlHostList = doc.createElementNS(ClusterMgmt.ClusterMgmtNS, ClusterMgmt.ClusterMgmtNSPrefix + ":HostList")
		cluster.appendChild(xmlHostList)
		
		hosts = self.getHostInfo()
		for host in hosts:
			xmlHost = doc.createElementNS(ClusterMgmt.ClusterMgmtNS, ClusterMgmt.ClusterMgmtNSPrefix + ":Host")
			xmlHost.setAttribute("name", host.HostName)
			xmlHostList.appendChild(xmlHost)			
			
			osName = doc.createElementNS(ClusterMgmt.ClusterMgmtNS, ClusterMgmt.ClusterMgmtNSPrefix + ":OSName")
			xmlHost.appendChild(osName)
			text = doc.createTextNode(host.OSName)
			osName.appendChild(text)
			
			cpuCount = doc.createElementNS(ClusterMgmt.ClusterMgmtNS, ClusterMgmt.ClusterMgmtNSPrefix + ":CpuCount")
			xmlHost.appendChild(cpuCount)			
			text = doc.createTextNode(str(host.CpuCount))
			cpuCount.appendChild(text)

			cpuArch = doc.createElementNS(ClusterMgmt.ClusterMgmtNS, ClusterMgmt.ClusterMgmtNSPrefix + ":CpuArch")
			xmlHost.appendChild(cpuArch)
			text = doc.createTextNode(str(host.CpuArch))
			cpuArch.appendChild(text)
			
			cpuSpeed = doc.createElementNS(ClusterMgmt.ClusterMgmtNS, ClusterMgmt.ClusterMgmtNSPrefix + ":CpuSpeed")
			xmlHost.appendChild(cpuSpeed)			
			text = doc.createTextNode(str(host.CpuSpeed))
			cpuSpeed.appendChild(text)

			additionalResources = doc.createElementNS(ClusterMgmt.ClusterMgmtNS, ClusterMgmt.ClusterMgmtNSPrefix + ":AdditionalResources")
			xmlHost.appendChild(additionalResources)
			for resource in host.AdditionalResources["lsf"]:
				additionalResource = doc.createElementNS(ClusterMgmt.ClusterMgmtNS, ClusterMgmt.ClusterMgmtNSPrefix + ":BooleanResource")
				additionalResource.setAttribute("id", str(resource))
				additionalResources.appendChild(additionalResource)
			
			freeFilesystem = doc.createElementNS(ClusterMgmt.ClusterMgmtNS, ClusterMgmt.ClusterMgmtNSPrefix + ":FreeFilesystem")
			xmlHost.appendChild(freeFilesystem)
			text = doc.createTextNode(str(int(1024 * 1024 * host.FreeFilesystem)))
			freeFilesystem.appendChild(text)
			
			freeMemory = doc.createElementNS(ClusterMgmt.ClusterMgmtNS, ClusterMgmt.ClusterMgmtNSPrefix + ":FreeMemory")
			xmlHost.appendChild(freeMemory)
			text = doc.createTextNode(str(int(1024 * 1024 * host.FreeMemory)))
			freeMemory.appendChild(text)
			
			freeSwap = doc.createElementNS(ClusterMgmt.ClusterMgmtNS, ClusterMgmt.ClusterMgmtNSPrefix + ":FreeSwap")
			xmlHost.appendChild(freeSwap)
			text = doc.createTextNode(str(int(host.FreeSwap * 1024 * 1024)))
			freeSwap.appendChild(text)

			jobNum = doc.createElementNS(ClusterMgmt.ClusterMgmtNS, ClusterMgmt.ClusterMgmtNSPrefix + ":JobNum")
			xmlHost.appendChild(jobNum)
			text = doc.createTextNode(str(host.JobNum))
			jobNum.appendChild(text)

			jobMax = doc.createElementNS(ClusterMgmt.ClusterMgmtNS, ClusterMgmt.ClusterMgmtNSPrefix + ":JobMax")
			xmlHost.appendChild(jobMax)
			text = doc.createTextNode(str(host.JobMax))
			jobMax.appendChild(text)


			freeSwap = doc.createElementNS(ClusterMgmt.ClusterMgmtNS, ClusterMgmt.ClusterMgmtNSPrefix + ":FreeSwap")
			xmlHost.appendChild(freeSwap)
			text = doc.createTextNode(str(int(host.FreeSwap * 1024 * 1024)))
			freeSwap.appendChild(text)


			if detailedHostList:
				cpuUtilization = doc.createElementNS(ClusterMgmt.ClusterMgmtNS, ClusterMgmt.ClusterMgmtNSPrefix + ":CpuUtilization")
				xmlHost.appendChild(cpuUtilization)			
				text = doc.createTextNode(str(host.CpuUtilization))
				cpuUtilization.appendChild(text)
				
				pagingActivity = doc.createElementNS(ClusterMgmt.ClusterMgmtNS, ClusterMgmt.ClusterMgmtNSPrefix + ":PagingActivity")
				xmlHost.appendChild(pagingActivity)
				text = doc.createTextNode(str(host.PagingActivity))
				pagingActivity.appendChild(text)
				
				queueLength = doc.createElementNS(ClusterMgmt.ClusterMgmtNS, ClusterMgmt.ClusterMgmtNSPrefix + ":QueueLength")
				queueLength.setAttribute("timespan", "900")
				xmlHost.appendChild(queueLength)
				text = doc.createTextNode(str(host.QueueLength15M))
				queueLength.appendChild(text)
				
				queueLength = doc.createElementNS(ClusterMgmt.ClusterMgmtNS, ClusterMgmt.ClusterMgmtNSPrefix + ":QueueLength")
				queueLength.setAttribute("timespan", "15")
				xmlHost.appendChild(queueLength)
				text = doc.createTextNode(str(host.QueueLength15S))
				queueLength.appendChild(text)
				
				queueLength = doc.createElementNS(ClusterMgmt.ClusterMgmtNS, ClusterMgmt.ClusterMgmtNSPrefix + ":QueueLength")
				queueLength.setAttribute("timespan", "60")
				xmlHost.appendChild(queueLength)
				text = doc.createTextNode(str(host.QueueLength1M))
				queueLength.appendChild(text)
				
				xmlJobList = doc.createElementNS(ClusterMgmt.ClusterMgmtNS, ClusterMgmt.ClusterMgmtNSPrefix + ":UndetailedJobList")
				xmlHost.appendChild(xmlJobList)
				for job in host.Jobs:
					xmlJob = doc.createElementNS(ClusterMgmt.ClusterMgmtNS, ClusterMgmt.ClusterMgmtNSPrefix + ":UndetailedJob")
					xmlJob.setAttribute("id", str(job))
					xmlJobList.appendChild(xmlJob)
					
				xmlQueueList = doc.createElementNS(ClusterMgmt.ClusterMgmtNS, ClusterMgmt.ClusterMgmtNSPrefix + ":UndetailedQueueList")
				xmlHost.appendChild(xmlQueueList)
				for queue in host.Queues:
					xmlQueue = doc.createElementNS(ClusterMgmt.ClusterMgmtNS, ClusterMgmt.ClusterMgmtNSPrefix + ":UndetailedQueue")
					xmlQueue.setAttribute("id", str(queue))
					xmlQueueList.appendChild(xmlQueue)
					
		xmlQueueList = doc.createElementNS(ClusterMgmt.ClusterMgmtNS, ClusterMgmt.ClusterMgmtNSPrefix + ":QueueList")
		cluster.appendChild(xmlQueueList)
		
		queues = self.getQueueInfo()
		for queue in queues:
			xmlQueue = doc.createElementNS(ClusterMgmt.ClusterMgmtNS, ClusterMgmt.ClusterMgmtNSPrefix + ":Queue")
			xmlQueue.setAttribute(ClusterMgmt.ClusterMgmtNSPrefix + ":Name", queue.QueueName)
			xmlQueueList.appendChild(xmlQueue)
			
			xmlJobList = doc.createElementNS(ClusterMgmt.ClusterMgmtNS, ClusterMgmt.ClusterMgmtNSPrefix + ":UndetailedJobList")
			xmlQueue.appendChild(xmlJobList)
			for job in queue.JobIDs:
				xmlJob = doc.createElementNS(ClusterMgmt.ClusterMgmtNS, ClusterMgmt.ClusterMgmtNSPrefix + ":UndetailedJob")
				xmlJob.setAttribute("id", str(job))
				xmlJobList.appendChild(xmlJob)
				
			maxJobs = doc.createElementNS(ClusterMgmt.ClusterMgmtNS, ClusterMgmt.ClusterMgmtNSPrefix + ":MaxJobs")			
			xmlQueue.appendChild(maxJobs)
			text = doc.createTextNode(str(queue.MaxJobs))
			maxJobs.appendChild(text)
			
			pendingJobs = doc.createElementNS(ClusterMgmt.ClusterMgmtNS, ClusterMgmt.ClusterMgmtNSPrefix + ":PendingJobs")
			xmlQueue.appendChild(pendingJobs)
			text = doc.createTextNode(str(queue.PendingJobs))
			pendingJobs.appendChild(text)

			runningJobs = doc.createElementNS(ClusterMgmt.ClusterMgmtNS, ClusterMgmt.ClusterMgmtNSPrefix + ":RunningJobs")
			xmlQueue.appendChild(runningJobs)
			text = doc.createTextNode(str(queue.RunningJobs))
			runningJobs.appendChild(text)
			
			queueEnabled = doc.createElementNS(ClusterMgmt.ClusterMgmtNS, ClusterMgmt.ClusterMgmtNSPrefix + ":QueueActive")
			xmlQueue.appendChild(queueEnabled)
			text = doc.createTextNode(str(queue.QueueEnabled))
			queueEnabled.appendChild(text)
			
			queuePriority = doc.createElementNS(ClusterMgmt.ClusterMgmtNS, ClusterMgmt.ClusterMgmtNSPrefix + ":QueuePriority")
			xmlQueue.appendChild(queuePriority)
			text = doc.createTextNode(str(queue.QueuePriority))
			queuePriority.appendChild(text)		
			
		
		xmlJobList = doc.createElementNS(ClusterMgmt.ClusterMgmtNS, ClusterMgmt.ClusterMgmtNSPrefix + ":JobList")
		cluster.appendChild(xmlJobList)
			
		jobs = self.getJobInfo()
		for job in jobs:
			xmlJob = doc.createElementNS(ClusterMgmt.ClusterMgmtNS, ClusterMgmt.ClusterMgmtNSPrefix + ":Job")
			xmlJob.setAttribute(ClusterMgmt.ClusterMgmtNSPrefix + ":Name", job.JobName)
			xmlJob.setAttribute(ClusterMgmt.ClusterMgmtNSPrefix + ":ID", str(job.JobID))
			xmlJobList.appendChild(xmlJob)

			jobStatus = doc.createElementNS(ClusterMgmt.ClusterMgmtNS, ClusterMgmt.ClusterMgmtNSPrefix + ":JobStatus")
			xmlJob.appendChild(jobStatus)
			text = doc.createTextNode(str(job.JobStatus))
			jobStatus.appendChild(text)
			
			if detailedJobList:			

				submissionTime = doc.createElementNS(ClusterMgmt.ClusterMgmtNS, ClusterMgmt.ClusterMgmtNSPrefix + ":SubmissionTime")
				xmlJob.appendChild(submissionTime)
				text = doc.createTextNode(str(job.SubmissionTime))
				submissionTime.appendChild(text)
				
				startTime = doc.createElementNS(ClusterMgmt.ClusterMgmtNS, ClusterMgmt.ClusterMgmtNSPrefix + ":StartTime")
				xmlJob.appendChild(startTime)
				text = doc.createTextNode(str(job.StartTime))
				startTime.appendChild(text)
				
				endTime = doc.createElementNS(ClusterMgmt.ClusterMgmtNS, ClusterMgmt.ClusterMgmtNSPrefix + ":EndTime")
				xmlJob.appendChild(endTime)
				text = doc.createTextNode(str(job.EndTime))
				endTime.appendChild(text)			
				
				systemTime = doc.createElementNS(ClusterMgmt.ClusterMgmtNS, ClusterMgmt.ClusterMgmtNSPrefix + ":SystemTime")
				xmlJob.appendChild(systemTime)
				text = doc.createTextNode(str(job.SystemTime))
				systemTime.appendChild(text)
				
				userTime = doc.createElementNS(ClusterMgmt.ClusterMgmtNS, ClusterMgmt.ClusterMgmtNSPrefix + ":UserTime")
				xmlJob.appendChild(userTime)
				text = doc.createTextNode(str(job.UserTime))
				userTime.appendChild(text)
				
				totalMem = doc.createElementNS(ClusterMgmt.ClusterMgmtNS, ClusterMgmt.ClusterMgmtNSPrefix + ":TotalMem")
				xmlJob.appendChild(totalMem)
				text = doc.createTextNode(str(job.TotalMem))
				totalMem.appendChild(text)
				
				totalSwap = doc.createElementNS(ClusterMgmt.ClusterMgmtNS, ClusterMgmt.ClusterMgmtNSPrefix + ":TotalSwap")
				xmlJob.appendChild(totalSwap)
				text = doc.createTextNode(str(job.TotalSwap))
				totalSwap.appendChild(text)
				
				xmlHostList = doc.createElementNS(ClusterMgmt.ClusterMgmtNS, ClusterMgmt.ClusterMgmtNSPrefix + ":UndetailedHostList")
				xmlJob.appendChild(xmlHostList)
				for host in job.ExecutionHosts:
					xmlHost = doc.createElementNS(ClusterMgmt.ClusterMgmtNS, ClusterMgmt.ClusterMgmtNSPrefix + ":UndetailedHost")
					xmlHost.setAttribute("name", str(host))
					xmlHostList.appendChild(xmlHost)
			
		return doc
