
#
# QosCosGrid Computing
# Copyright (C) 2008-2011 Poznan Supercomputing and Networking Center
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>
#

import datetime, sys, subprocess, re, traceback, time, os
from common import HostInfo, QueueInfo, JobInfo, ReservationInfo, asList, asDict 
from gmexceptions import GMInternalError, GMExternalError, GMUnknowJob

__statusRegex = re.compile(r'([^=]+)=([^,]*),*')
__queueStateRegex = re.compile(r'([^:]+):([^ ]*)[ ]*')
__execHostRegex = re.compile(r'([^/]+)/\d+(\*(\d+))?\+*')
__valueRegex = re.compile(r'([\d\.]+)(\D*)')
__walltimeRegex = re.compile(r'([0-9]+):([0-9]+):([0-9]+)')
__resACLRegex = re.compile(r'^\s*ACL:.*USER==(\w+)', re.MULTILINE)
__resAccountingRegex = re.compile(r'^\s*Accounting.*User:(\w+)', re.MULTILINE)
__newlineRegex = re.compile(r'\n')
__emptyLineRegex = re.compile(r'\n[\s\n]*\n')
__equalRegex = re.compile(r'[ ]+=[ ]+')
__jobIdRegex = re.compile(r'[^\d]*([\d\-]+.*)')
__jobSplitRegex = re.compile(r'\nJob Id')
__queueRegex = re.compile(r'Queue: (.+)')
__hostsInfoDict = None
__hostsInfoList = None
__jobsInfoDict = None
__jobsInfoList = None
__queuesInfoList = None
__queuesInfoDict = None

def __checknodeResRegexExp(value):
    return "User \'" + value + "\'.*Procs: (\d+)/(\d+)" 

__defaultOwner = "root"
__checknodeROI = "Reservations:"

__postfixTable = {
    "": 1,
    "kb": 1024,
    "mb": 1024*1024,
    "gb": 1024*1024*1024
}

def __extractPostFix(value):
    m = __valueRegex.match(value)
    ret = [m.group(1)]
    if m.group(2)!=None:
        ret.append(m.group(2))
    else:
        ret.append("")
    return ret

def __calculateValue(value):
    p = __extractPostFix(value)
    return float(round(float(p[0]) \
        *__postfixTable.get(p[1]))) 

def __commandOutputToDict(output, items_list=None, key_regex=None, split_regex=None):
    if split_regex == None:
        blocks_list = __emptyLineRegex.split(output)
    else:
        blocks_list = split_regex.split(output)
    items_dictionary = {}
    for block in blocks_list:
        if block.lstrip() == "":
            continue
        block_lines = __newlineRegex.split(block)

        if key_regex != None:
            block_name = key_regex.match(block_lines[0]).group(1)
        else:
            block_name = block_lines[0]

        block_lines = block_lines[1:]
        block_dict = {}
        for line in block_lines:
            tuple = __equalRegex.split(line, 1)
            if len(tuple) == 2:
                block_dict[ tuple[0].strip() ] = tuple[1].strip()

        if items_list != None:
            items_list.append(block_name)

        items_dictionary[block_name] = block_dict
     
    return items_dictionary

def getHostData():
    global __hostsInfoDict
    global __hostsInfoList
    if __hostsInfoList is not None:
        return (__hostsInfoList, __hostsInfoDict)
 
    hostsInfoDict = dict()
    hostsInfoList = []
    pbsnodes_output = _pbsnodes(['-a'])
    nodes_property = os.getenv("QCG_NODE_PROPERTY_ATTR")
    nodes_filter = os.getenv("QCG_NODE_FILTER")
    if nodes_filter != None:
        try:
            nodes_filter_key = nodes_filter.split(":")[0]
            nodes_filter_regex_str = nodes_filter.split(":")[1]
            if nodes_filter_regex_str[:1] == '~':
                nodes_filter_negative = True
                nodes_filter_regex_str = nodes_filter_regex_str[1:]
            else:
                nodes_filter_negative = False
            nodes_filter_regex = re.compile(nodes_filter_regex_str)
        except Exception:
            raise GMInternalError("QCG_NODE_FILTER: Invalid format: '" + nodes_filter + "':. Expected: 'field:regex'")

    pbsnodes_nodes_list = __emptyLineRegex.split(pbsnodes_output)
    nodes_list = []
    nodes_dictionary = __commandOutputToDict(pbsnodes_output, nodes_list)

    for node_name in nodes_list:
        current_host = HostInfo()
        current_host.HostName = node_name;

        node_properties = nodes_dictionary[node_name];
        if nodes_filter != None:
            if nodes_filter_key not in node_properties:
                current_host.Dedicated = True
            if (nodes_filter_regex.search(node_properties[nodes_filter_key]) == None) ^ nodes_filter_negative:
                current_host.Dedicated = True

        if node_properties['state'].find("down") != -1: 
            current_host.State = "down"
        elif node_properties['state'].find("offline") != -1:
            current_host.State = "offline"
        elif node_properties['state'].find("state-unknown") != -1:
            current_host.State = "down"
        else:
            current_host.State = "up"

        if nodes_property is not None and nodes_property in node_properties:
            current_host.AdditionalResources = node_properties[nodes_property]

        current_host.CpuArch = "x86_64"
    
        try:
            current_host.TotalMemory = __calculateValue(node_properties["resources_available.mem"])
            current_host.CpuCount = int(node_properties['resources_available.ncpus'])
            current_host.UsedCpuCount = int(node_properties['resources_assigned.ncpus'])
            current_host.TotalSwap = 0.0
            hostsInfoDict[node_name] = current_host
            hostsInfoList.append(current_host)
        except:
            pass

    return (hostsInfoList, hostsInfoDict)

def __decodeHostStatusField(status):
    ret = {} 
    for m in __statusRegex.finditer(status):
        ret[m.group(1)] = m.group(2)        
    return ret

def __decodeQueueStateField(state):
    ret = {}
    for m in __queueStateRegex.finditer(state):
        ret[m.group(1)] = m.group(2)
    return ret

def __decodeTimeDelta(time):
    match = __walltimeRegex.match(time)
    sec = int(match.group(1)) * 3600 + int(match.group(2)) * 60 + int(match.group(3))
    return datetime.timedelta(seconds=sec) 
        
def getJobData():
    global __jobsInfoList 
    global __jobsInfoDict 
      
    if __jobsInfoList is not None: 
         return (__jobsInfoList,__jobsInfoDict) 

    __jobsInfoDict = dict()
    __jobsInfoList = []
    qstat_output = _qstat(['-f'])
    job_list = __commandOutputToDict(qstat_output, None, __jobIdRegex, __jobSplitRegex)
    for jobid,job_info in job_list.iteritems():
        decoded_job = __decodeJobData(jobid, job_info)
        __jobsInfoList.append(decoded_job)
        __jobsInfoDict[jobid] = decoded_job

    return (__jobsInfoList,__jobsInfoDict) 

def __decodeJobData(jobid, attributes):
    current_job = JobInfo()
    current_job.JobID = jobid
    current_job.JobName = attributes["Job_Name"]
    if attributes["job_state"] == "R" or attributes["job_state"] == "E":
        current_job.JobStatus = "Running"
    elif attributes["job_state"] == "Q":
        current_job.JobStatus = "Pending"
    elif attributes["job_state"] == "C":
        current_job.JobStatus = "Finished"
    else:
        current_job.JobStatus = "Unknown"
    current_job.OwnerName = attributes["Job_Owner"].split("@")[0]
    current_job.QueueName = attributes["queue"]
    current_job.SubmissionTime = datetime.datetime(*time.strptime(attributes["qtime"])[:6])
    current_job.TotalCores = int(attributes["Resource_List.ncpus"])

    #TODO walltime, cputime, estimated time, comment
    if "resources_used.mem" in attributes:
        current_job.MemUsage =  long(__calculateValue(attributes["resources_used.mem"]) / (1024*1024))
    if "resources_used.cput" in attributes:
        current_job.CpuTime = __decodeTimeDelta(attributes["resources_used.cput"]) 
    if "resources_used.walltime" in attributes: 
        current_job.Walltime = __decodeTimeDelta(attributes["resources_used.walltime"]) 
    if "comment" in attributes:
        current_job.Comment = attributes["comment"]

    if attributes["job_state"] == "Q":
        try:
            current_job.EstStartTime = datetime.datetime(*time.strptime(attributes["estimated.start_time"])[:6])
        except:
            pass
    else:
        current_job.StartTime = datetime.datetime(*time.strptime(attributes["mtime"])[:6])
    return current_job

def addExtendedJobData(job_ids, all_job_ids):
    qstat_output = _qstat(['-n', '-1']).split("\n") #there is no huge difference if we want to parse all or just selected jobs
    for line in qstat_output:
        if len(line) > 0 and line[0].isdigit():
            columns = line.split()
            job_id = columns[0]
            exec_host = columns[11]

            for m in __execHostRegex.finditer(exec_host): 
                host_name = m.group(1)
                if m.group(3) != None:
                    __jobsInfoDict[job_id].ExecutionHosts[host_name] = int(m.group(3))
                else:
                    __jobsInfoDict[job_id].ExecutionHosts[host_name] = 1 
   
    pass #PBS Pro is so cute that here we need just to get exec_hosts
    
def getQueueData():
    global __queuesInfoList 
    global __queuesInfoDict 

    if __queuesInfoList is not None: 
        return (__queuesInfoList, __queuesInfoDict) 

    #TODO try to obtain default queue name
    default_queue = None
    qstat_output = _qstat([ '-f', '-Q'])
    all_queues = __commandOutputToDict(qstat_output, None, __queueRegex)
    __queuesInfoDict = dict()  
    __queuesInfoList = []
    for queue_name, queue_dict in all_queues.iteritems():
            decodedQueue = __decodeQueueData(queue_name, queue_dict) 
            if decodedQueue.QueueName == default_queue:
                decodedQueue.IsDefault = True
            __queuesInfoDict[queue_name] = decodedQueue
            __queuesInfoList.append(decodedQueue)     

    return (__queuesInfoList, __queuesInfoDict)

def __decodeQueueData(queue_name, attributes):
    current_queue = QueueInfo()
    current_queue.QueueName = queue_name
    queue_state = __decodeQueueStateField(attributes["state_count"])    
    current_queue.PendingJobs = int(queue_state["Queued"])  
    current_queue.RunningJobs = int(queue_state["Running"])
    if "resources_default.walltime" in attributes.keys():
        current_queue.Walltime = __decodeTimeDelta(attributes["resources_default.walltime"])
    elif "resources_max.walltime" in attributes.keys():
        current_queue.Walltime = __decodeTimeDelta(attributes["resources_max.walltime"])
    else:
        current_queue.Walltime = datetime.timedelta(0)  
    current_queue.job_slots_aval = 0
    current_queue.job_slots_total = 0
    current_queue.Enabled = (attributes["enabled"] == "True")
    current_queue.Started = (attributes["started"] == "True")
 
    return current_queue

def _qstat(args):
    try:
        command = ['qstat']
        command.extend(args)
        sr = subprocess.Popen(command, stdout=subprocess.PIPE)
    except OSError, e:
        raise GMExternalError("qstat command failed, reason:\n" + str(e))        
    
    qstat_output = sr.communicate()[0]
    if sr.returncode != 0:
        if sr.returncode == 153:
            raise GMUnknowJob("Unknow Job")
        else:
            raise GMExternalError("qstat command returned code: " + str(sr.returncode) + ". Details: " + qstat_output)
    return qstat_output


def _pbsnodes(args):
    try:
        command = ['pbsnodes']
        command.extend(args)
        sr = subprocess.Popen(command, stdout=subprocess.PIPE)
    except OSError, e:
        raise GMExternalError("pbsnodes command failed, reason:\n" + str(e))        
    
    pbsnodes_output = sr.communicate()[0]
    if sr.returncode != 0:
        raise GMExternalError("pbsnodes command returned code: " + str(sr.returncode))
    return pbsnodes_output

def getReservationData():
    return ([], dict())

