# $Id$

#
# QosCosGrid Computing
# Copyright (C) 2008-2011 Poznan Supercomputing and Networking Center
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>
#

import os
import re
from xml.dom import minidom
from common import HostInfo, QueueInfo, JobInfo, GridMonNSPrefix, GridMonNS, BesFactoryNSPrefix, BesFactoryNS, JsdlNSPrefix, JsdlNS
from datetime import datetime
from xml.dom.minidom import Document

__valueRegex = re.compile(r'([\d\.]+)(\D*)')
__archRegex = re.compile(r'(.*)-(.*)')
__boolExp = reg = re.compile(r'(\w+)\s+\w+\s+BOOL.+')

__postfixTable = {
	"": 1,
	"K": 1024,
	"M": 1024 * 1024,
	"G": 1024 * 1024 * 1024
}

__systemTable = {
	"lx24": "LINUX",
	"lx26": "LINUX"
}

__stateTable = {
	"qw": "Pending",
	"r": "Running",
	"w": "Suspended"	
}

def __extractPostFix(value):
	m = __valueRegex.match(value)
	ret = [m.group(1)]
	if m.group(2) != None:
		ret.append(m.group(2))
	else:
		ret.append("")
	return ret

def __calculateValue(value):
	p = __extractPostFix(value)
	return float(round(float(p[0]) \
		* __postfixTable.get(p[1])))


def __handleArch(hostinfo, value):
	m = __archRegex.match(value)
	if m.group(1) != "":
		hostinfo.OSName = __systemTable[m.group(1)]
	else:
		hostinfo.OSName = "Unknown"
	if m.group(2) != "":
		hostinfo.CpuArch = str(m.group(2))
	else:
		hostinfo.CpuArch = "Other"

def __handleCpuCount(hostinfo, value):
	if value != '-':
		hostinfo.CpuCount = int(value)
	else:
		hostinfo.CpuCount = None

def __handleLoadAvg(hostinfo, value):
	if value != '-':
		hostinfo.CpuUtilization = float(value)	

def __handleMemTotal(hostinfo, value):
	if value != '-':
		hostinfo.total_mem = __calculateValue(value)
	else:
		hostinfo.total_mem = None

def __handleMemUsed(hostinfo, value):
	if value != '-' and hostinfo.total_mem != None:
		hostinfo.FreeMemory = hostinfo.total_mem - __calculateValue(value)	

def __handleSwapTotal(hostinfo, value):
	if value != '-':
		hostinfo.total_swap = __calculateValue(value)
	else:
		hostinfo.total_swap = None

def __handleSwapUsed(hostinfo, value):
	if value != '-' and hostinfo.total_swap != None:
		hostinfo.FreeSwap = hostinfo.total_swap - __calculateValue(value)	

def __handleOsName(hostinfo, value, name):
	hostinfo.OSName = str(value)

def __handleCpuSpeed(hostinfo, value, name):
	hostinfo.CpuSpeed = float(value)

def __handleCpuArch(hostinfo, value, name):
	hostinfo.CpuArch = str(value)

def __handleBoolRes(hostinfo, value, name):
	hostinfo.AdditionalResources["sge"].append(name)
		
def __getValueFromNode(node, tagname):
		tag = node.getElementsByTagName(tagname)
		if len(tag) > 0 and len(tag[0].childNodes) > 0:
			return tag[0].childNodes[0].nodeValue
		return ""

__handleQHostNode = {
	"arch_string": __handleArch,
	"num_proc": __handleCpuCount,
	"load_avg": __handleLoadAvg,
	"mem_total": __handleMemTotal,
	"mem_used": __handleMemUsed,
	"swap_total": __handleSwapTotal,
	"swap_used": __handleSwapUsed
}

__handleQHostResource = {
	"osname": __handleOsName,
	"cpuarch": __handleCpuArch,
	"cpuspeed": __handleCpuSpeed
}

#qconf_command = "qconf -sc"
#out = os.popen(qconf_command)
#line = out.readline()
#while(line != ""):
#	m = __boolExp.match(line)
#	if m != None:
#		__handleQHostResource[m.group(1)] = __handleBoolRes
#	line = out.readline()


		
def getHostData(hostnames=None):
			
	command = "qhost -j -F -q -xml"
	hostsInfo = []

	if hostnames != None :
		command += " -h"
		for name in hostnames:
			command += " " + name
	out = os.popen(command)
	dom = minidom.parse(out)
	hosts = dom.getElementsByTagName('host')
	for host in hosts:
		current_host = HostInfo()
		current_host.AdditionalResources["sge"] = []
		current_host.HostName = str(host.getAttribute('name'))
		if current_host.HostName == "global":
			continue
		for host_info in host.getElementsByTagName('hostvalue'):
			property = host_info.getAttribute('name')
			value = host_info.childNodes[0].nodeValue
			if __handleQHostNode.get(property) != None:
				__handleQHostNode.get(property)(current_host, value)

		for host_info in host.getElementsByTagName('resourcevalue'):
			property = host_info.getAttribute('name')
			value = host_info.childNodes[0].nodeValue
			if __handleQHostResource.get(property) != None:
				__handleQHostResource.get(property)(current_host, value, property)

		for queue_info in host.getElementsByTagName('queue'):
			current_host.Queues.append(queue_info.getAttribute('name'))
		for job_info in host.getElementsByTagName('job'):
			current_host.Jobs.append(job_info.getAttribute('name'))
		hostallq = getQueueData(["all.q@" + current_host.HostName])
		current_host.JobMax = hostallq[0].MaxJobs
		current_host.JobNum = hostallq[0].PendingJobs
		current_host.FreeFilesystem = 0.0
		hostsInfo.append(current_host)
	return hostsInfo

def getQueueData(queuenames=None):
	command = "qstat -f -xml"
	queuesInfo = []

	if queuenames != None :
		command += " -q"
		for queue in queuenames:
			command += " " + queue
	out = os.popen(command)
	dom = minidom.parse(out)
	dom = dom.childNodes[0]
	queues = dom.getElementsByTagName('Queue-List')
	for queue in queues:
		current_queue = QueueInfo()		
		current_queue.QueueName = str(__getValueFromNode(queue, 'name'))
		current_queue.MaxJobs = int(__getValueFromNode(queue, 'slots_total'))
		current_queue.job_slots_aval = current_queue.MaxJobs - \
									int(__getValueFromNode(queue, 'slots_used'))
		current_queue.PendingJobs = current_queue.MaxJobs - current_queue.job_slots_aval
		current_queue.queue_state = __getValueFromNode(queue, 'state')
		for job in queue.getElementsByTagName('job_list'):
			current_queue.JobIDs.append(int(__getValueFromNode(job, 'JB_job_number')))
		#retrive pending jobs
		for node in dom.childNodes:
			if node.localName == "job_info":
				for job in node.getElementsByTagName('job_list'):
					current_queue.JobIDs.append(int(__getValueFromNode(job, 'JB_job_number')))
		queuesInfo.append(current_queue)
	return queuesInfo

def getJobData(jobnames=None):
	command = "qstat -xml"
	jobsInfo = []

	out = os.popen(command)
	dom = minidom.parse(out)

	jobs = dom.getElementsByTagName('job_list')
	for job in jobs:
		current_job = JobInfo()
		current_job.JobID = int(__getValueFromNode(job, 'JB_job_number'))
		if jobnames != None and not current_job.job_id in jobnames:
			break
		current_job.JobName = str(__getValueFromNode(job, 'JB_name'))
		current_job.JobStatus = __stateTable[str(__getValueFromNode(job, 'state'))]
		current_job.OwnerName = str(__getValueFromNode(job, 'JB_owner'))
		current_job.QueueName = str(__getValueFromNode(job, 'queue_name'))
		current_job.SubmissionTime = datetime.strptime(__getValueFromNode(job, 'JB_submission_time'), "%Y-%m-%dT%H:%M:%S")
		start_time = __getValueFromNode(job, 'JAT_start_time')
		if start_time != "":
			current_job.StartTime = datetime.strptime(start_time, "%Y-%m-%dT%H:%M:%S")
		jobsInfo.append(current_job)
	return jobsInfo

def getReservationData(reservation_names=None):
	reservations_info = []
	return reservations_info

def besXml():
	"""Gets information about hosts in XML format (bes-factory namespace)."""
	doc = Document()

	cluster = doc.createElementNS(GridMonNS, GridMonNSPrefix + ":Cluster")
	cluster.setAttribute("type", "SGE")
	doc.appendChild(cluster)		

	hosts = getHostData()
	for host in hosts:
		containedResource = doc.createElement("ContainedResource")
		cluster.appendChild(containedResource)

		resourceName = doc.createElementNS(BesFactoryNS, BesFactoryNSPrefix + ":ResourceName")
		containedResource.appendChild(resourceName)
		text = doc.createTextNode(str(host.HostName))
		resourceName.appendChild(text)

		operatingSystem = doc.createElementNS(BesFactoryNS, BesFactoryNSPrefix + ":OperatingSystem")
		containedResource.appendChild(operatingSystem)
		operatingSystemType = doc.createElementNS(JsdlNS, JsdlNSPrefix + ":OperatingSystemType")
		operatingSystem.appendChild(operatingSystemType)
		operatingSystemName = doc.createElementNS(JsdlNS, JsdlNSPrefix + ":OperatingSystemName")
		operatingSystemType.appendChild(operatingSystemName)
		text = doc.createTextNode(str(host.OSName))
		operatingSystemName.appendChild(text)

		cpuArchitecture = doc.createElementNS(BesFactoryNS, BesFactoryNSPrefix + ":CPUArchitecture")
		containedResource.appendChild(cpuArchitecture)
		cpuArchitectureName = doc.createElementNS(JsdlNS, JsdlNSPrefix + ":CPUArchitectureName")
		cpuArchitecture.appendChild(cpuArchitectureName)
		text = doc.createTextNode(str(host.CpuArch))
		cpuArchitectureName.appendChild(text)

		cpuCount = doc.createElementNS(BesFactoryNS, BesFactoryNSPrefix + ":CPUCount")
		containedResource.appendChild(cpuCount)
		text = doc.createTextNode(str(int(host.CpuCount)))
		cpuCount.appendChild(text)

		cpuSpeed = doc.createElementNS(BesFactoryNS, BesFactoryNSPrefix + ":CPUSpeed")
		containedResource.appendChild(cpuSpeed)
		text = doc.createTextNode(str(int(host.CpuSpeed * 1024 * 1024)))
		cpuSpeed.appendChild(text)

	return doc
	

def getAllInfoXml(detailedHostList=False, detailedJobList=False):
	"""Gets information about this cluster in XML format.
	See schema for format details.
	"""
	doc = Document()	
	
	cluster = doc.createElementNS(GridMonNS, GridMonNSPrefix + ":Cluster")
	cluster.setAttribute("type", "SGE")
	doc.appendChild(cluster)
	
	xmlHostList = doc.createElementNS(GridMonNS, GridMonNSPrefix + ":HostList")
	cluster.appendChild(xmlHostList)
	
	hosts = getHostData()
	for host in hosts:
		xmlHost = doc.createElementNS(GridMonNS, GridMonNSPrefix + ":Host")
		xmlHost.setAttribute("name", host.HostName)
		xmlHostList.appendChild(xmlHost)			
		
		osName = doc.createElementNS(GridMonNS, GridMonNSPrefix + ":OSName")
		xmlHost.appendChild(osName)
		text = doc.createTextNode(host.OSName)
		osName.appendChild(text)
		
		cpuCount = doc.createElementNS(GridMonNS, GridMonNSPrefix + ":CpuCount")
		xmlHost.appendChild(cpuCount)			
		text = doc.createTextNode(str(host.CpuCount))
		cpuCount.appendChild(text)

		cpuArch = doc.createElementNS(GridMonNS, GridMonNSPrefix + ":CpuArch")
		xmlHost.appendChild(cpuArch)
		text = doc.createTextNode(str(host.CpuArch))
		cpuArch.appendChild(text)
		
		cpuSpeed = doc.createElementNS(GridMonNS, GridMonNSPrefix + ":CpuSpeed")
		xmlHost.appendChild(cpuSpeed)			
		text = doc.createTextNode(str(host.CpuSpeed))
		cpuSpeed.appendChild(text)

		additionalResources = doc.createElementNS(GridMonNS, GridMonNSPrefix + ":AdditionalResources")
		xmlHost.appendChild(additionalResources)
		for resource in host.AdditionalResources["sge"]:
			additionalResource = doc.createElementNS(GridMonNS, GridMonNSPrefix + ":BooleanResource")
			additionalResource.setAttribute("id", str(resource))
			additionalResources.appendChild(additionalResource)
		
		freeFilesystem = doc.createElementNS(GridMonNS, GridMonNSPrefix + ":FreeFilesystem")
		xmlHost.appendChild(freeFilesystem)
		text = doc.createTextNode(str(int(1024 * 1024 * host.FreeFilesystem)))
		freeFilesystem.appendChild(text)
		
		freeMemory = doc.createElementNS(GridMonNS, GridMonNSPrefix + ":FreeMemory")
		xmlHost.appendChild(freeMemory)
		text = doc.createTextNode(str(host.FreeMemory))
		freeMemory.appendChild(text)
		
		freeSwap = doc.createElementNS(GridMonNS, GridMonNSPrefix + ":FreeSwap")
		xmlHost.appendChild(freeSwap)
		text = doc.createTextNode(str(host.FreeSwap))
		freeSwap.appendChild(text)

		jobNum = doc.createElementNS(GridMonNS, GridMonNSPrefix + ":JobNum")
		xmlHost.appendChild(jobNum)
		text = doc.createTextNode(str(host.JobNum))
		jobNum.appendChild(text)

		jobMax = doc.createElementNS(GridMonNS, GridMonNSPrefix + ":JobMax")
		xmlHost.appendChild(jobMax)
		text = doc.createTextNode(str(host.JobMax))
		jobMax.appendChild(text)


		freeSwap = doc.createElementNS(GridMonNS, GridMonNSPrefix + ":FreeSwap")
		xmlHost.appendChild(freeSwap)
		text = doc.createTextNode(str(host.FreeSwap))
		freeSwap.appendChild(text)


		if detailedHostList:
			cpuUtilization = doc.createElementNS(GridMonNS, GridMonNSPrefix + ":CpuUtilization")
			xmlHost.appendChild(cpuUtilization)			
			text = doc.createTextNode(str(host.CpuUtilization))
			cpuUtilization.appendChild(text)
			
			pagingActivity = doc.createElementNS(GridMonNS, GridMonNSPrefix + ":PagingActivity")
			xmlHost.appendChild(pagingActivity)
			text = doc.createTextNode(str(host.PagingActivity))
			pagingActivity.appendChild(text)
			
			queueLength = doc.createElementNS(GridMonNS, GridMonNSPrefix + ":QueueLength")
			queueLength.setAttribute("timespan", "900")
			xmlHost.appendChild(queueLength)
			text = doc.createTextNode(str(host.QueueLength15M))
			queueLength.appendChild(text)
			
			queueLength = doc.createElementNS(GridMonNS, GridMonNSPrefix + ":QueueLength")
			queueLength.setAttribute("timespan", "15")
			xmlHost.appendChild(queueLength)
			text = doc.createTextNode(str(host.QueueLength15S))
			queueLength.appendChild(text)
			
			queueLength = doc.createElementNS(GridMonNS, GridMonNSPrefix + ":QueueLength")
			queueLength.setAttribute("timespan", "60")
			xmlHost.appendChild(queueLength)
			text = doc.createTextNode(str(host.QueueLength1M))
			queueLength.appendChild(text)
			
			xmlJobList = doc.createElementNS(GridMonNS, GridMonNSPrefix + ":UndetailedJobList")
			xmlHost.appendChild(xmlJobList)
			for job in host.Jobs:
				xmlJob = doc.createElementNS(GridMonNS, GridMonNSPrefix + ":UndetailedJob")
				xmlJob.setAttribute("id", str(job))
				xmlJobList.appendChild(xmlJob)
				
			xmlQueueList = doc.createElementNS(GridMonNS, GridMonNSPrefix + ":UndetailedQueueList")
			xmlHost.appendChild(xmlQueueList)
			for queue in host.Queues:
				xmlQueue = doc.createElementNS(GridMonNS, GridMonNSPrefix + ":UndetailedQueue")
				xmlQueue.setAttribute("id", str(queue))
				xmlQueueList.appendChild(xmlQueue)
				
	xmlQueueList = doc.createElementNS(GridMonNS, GridMonNSPrefix + ":QueueList")
	cluster.appendChild(xmlQueueList)
	
	queues = getQueueData()
	for queue in queues:
		xmlQueue = doc.createElementNS(GridMonNS, GridMonNSPrefix + ":Queue")
		xmlQueue.setAttribute(GridMonNSPrefix + ":Name", queue.QueueName)
		xmlQueueList.appendChild(xmlQueue)
		
		xmlJobList = doc.createElementNS(GridMonNS, GridMonNSPrefix + ":UndetailedJobList")
		xmlQueue.appendChild(xmlJobList)
		for job in queue.JobIDs:
			xmlJob = doc.createElementNS(GridMonNS, GridMonNSPrefix + ":UndetailedJob")
			xmlJob.setAttribute("id", str(job))
			xmlJobList.appendChild(xmlJob)
			
		maxJobs = doc.createElementNS(GridMonNS, GridMonNSPrefix + ":MaxJobs")			
		xmlQueue.appendChild(maxJobs)
		text = doc.createTextNode(str(queue.MaxJobs))
		maxJobs.appendChild(text)
		
		pendingJobs = doc.createElementNS(GridMonNS, GridMonNSPrefix + ":PendingJobs")
		xmlQueue.appendChild(pendingJobs)
		text = doc.createTextNode(str(queue.PendingJobs))
		pendingJobs.appendChild(text)

		runningJobs = doc.createElementNS(GridMonNS, GridMonNSPrefix + ":RunningJobs")
		xmlQueue.appendChild(runningJobs)
		text = doc.createTextNode(str(queue.RunningJobs))
		runningJobs.appendChild(text)
		
		queueEnabled = doc.createElementNS(GridMonNS, GridMonNSPrefix + ":QueueActive")
		xmlQueue.appendChild(queueEnabled)
		text = doc.createTextNode(str(queue.QueueEnabled))
		queueEnabled.appendChild(text)
		
		queuePriority = doc.createElementNS(GridMonNS, GridMonNSPrefix + ":QueuePriority")
		xmlQueue.appendChild(queuePriority)
		text = doc.createTextNode(str(queue.QueuePriority))
		queuePriority.appendChild(text)		
		
	
	xmlJobList = doc.createElementNS(GridMonNS, GridMonNSPrefix + ":JobList")
	cluster.appendChild(xmlJobList)
		
	jobs = getJobData()
	for job in jobs:
		xmlJob = doc.createElementNS(GridMonNS, GridMonNSPrefix + ":Job")
		xmlJob.setAttribute(GridMonNSPrefix + ":Name", job.JobName)
		xmlJob.setAttribute(GridMonNSPrefix + ":ID", str(job.JobID))
		xmlJobList.appendChild(xmlJob)

		jobStatus = doc.createElementNS(GridMonNS, GridMonNSPrefix + ":JobStatus")
		xmlJob.appendChild(jobStatus)
		text = doc.createTextNode(str(job.JobStatus))
		jobStatus.appendChild(text)
		
		if detailedJobList:			

			submissionTime = doc.createElementNS(GridMonNS, GridMonNSPrefix + ":SubmissionTime")
			xmlJob.appendChild(submissionTime)
			text = doc.createTextNode(str(job.SubmissionTime))
			submissionTime.appendChild(text)
			
			startTime = doc.createElementNS(GridMonNS, GridMonNSPrefix + ":StartTime")
			xmlJob.appendChild(startTime)
			text = doc.createTextNode(str(job.StartTime))
			startTime.appendChild(text)
			
			endTime = doc.createElementNS(GridMonNS, GridMonNSPrefix + ":EndTime")
			xmlJob.appendChild(endTime)
			text = doc.createTextNode(str(job.EndTime))
			endTime.appendChild(text)			
			
			systemTime = doc.createElementNS(GridMonNS, GridMonNSPrefix + ":SystemTime")
			xmlJob.appendChild(systemTime)
			text = doc.createTextNode(str(job.SystemTime))
			systemTime.appendChild(text)
			
			userTime = doc.createElementNS(GridMonNS, GridMonNSPrefix + ":UserTime")
			xmlJob.appendChild(userTime)
			text = doc.createTextNode(str(job.UserTime))
			userTime.appendChild(text)
			
			totalMem = doc.createElementNS(GridMonNS, GridMonNSPrefix + ":TotalMem")
			xmlJob.appendChild(totalMem)
			text = doc.createTextNode(str(job.TotalMem))
			totalMem.appendChild(text)
			
			totalSwap = doc.createElementNS(GridMonNS, GridMonNSPrefix + ":TotalSwap")
			xmlJob.appendChild(totalSwap)
			text = doc.createTextNode(str(job.TotalSwap))
			totalSwap.appendChild(text)
			
			xmlHostList = doc.createElementNS(GridMonNS, GridMonNSPrefix + ":UndetailedHostList")
			xmlJob.appendChild(xmlHostList)
			for host in job.ExecutionHosts:
				xmlHost = doc.createElementNS(GridMonNS, GridMonNSPrefix + ":UndetailedHost")
				xmlHost.setAttribute("name", str(host))
				xmlHostList.appendChild(xmlHost)
		
	return doc
