
# $Id$

#
# QosCosGrid Computing
# Copyright(C) 2008-2011 Poznan Supercomputing and Networking Center
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
#(at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.    See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.    If not, see <http://www.gnu.org/licenses/>
#

try:
        from qcg.core import *
except ImportError:
        def error(msg): 
                print >> sys.stderr, "[ERROR] " + msg
        def info(msg): 
                print >> sys.stderr, "[INFO] " + msg


import datetime, sys, subprocess, re, traceback, time
from common import HostInfo, QueueInfo, JobInfo, ReservationInfo
from gmexceptions import GMInternalError, GMExternalError, GMUnknowJob

__hostsInfoDict = None 
__hostsInfoList = None 
__jobsInfoDict = None 
__jobsInfoList = None 
__queuesInfoList = None 
__queuesInfoDict = None 
__reservationsInfoDict = None 
__reservationsInfoList = None 
__hostnamesDict = dict()

def _runCommand(command, args):
    try:
        commandList = [command]
        commandList.extend(args)
        sr = subprocess.Popen(commandList, stdout=subprocess.PIPE)
    except OSError:
        raise GMExternalError(command + " command failed, reason:\n" + traceback.format_exc())

    commandOutput = sr.communicate()[0]
    if sr.returncode != 0:
        raise GMExternalError(command + "    command returned code: " + str(sr.returncode) + ". Details: " + commandOutput)
    return commandOutput

def __decodeDatetime(strTime):
    return datetime.datetime(*(time.strptime(strTime, '%Y-%m-%dT%H:%M:%S')[0:6]))

def __decodeHostnames(hostnames):
    if hostnames not in __hostnamesDict:
        __hostnamesDict[hostnames] = _runCommand('scontrol', ['show', 'hostnames', hostnames]).splitlines()
    return __hostnamesDict[hostnames]

def __formatOutputDict(output, items_list=None, key_regex=None):

    outputList = []

    for line in output.splitlines():
        propertiesDict = {}

        for prop in line.split():
            if not '=' in prop:
                continue

            name, value = prop.split('=', 1)

            if value == '(null)':
                value = None

            propertiesDict[name] = value

        outputList.append(propertiesDict)

    return outputList

def getHostData():
    global __hostsInfoDict 
    global __hostsInfoList 

    if __hostsInfoList is not None: 
        return(__hostsInfoList, __hostsInfoDict) 
         
    __hostsInfoDict = dict() 
    __hostsInfoList = [] 

    queueOutput = _runCommand('squeue', ['--states=r', '-h', '--format=%i %N'])
    jobs = __decodeNodeJobs(queueOutput)

    showPartitionOutput = _runCommand('scontrol', ['show', 'partition', '-o'])
    partition = __decodePartitionNodes(showPartitionOutput)

    showNodeOutput = _runCommand('scontrol', ['show', 'node', '-o'])
    nodeDescList = __formatOutputDict(showNodeOutput)


    for node in nodeDescList:
        currentNode = HostInfo()
        currentNode.HostName = node['NodeName']
        currentNode.CpuCount = int(node['CPUTot'])
        currentNode.UsedCpuCount = int(node['CPUAlloc'])
        currentNode.TotalMemory = float(node['RealMemory']) * 2**20
        if 'Features' in node and node['Features'] is not None:
            currentNode.AdditionalResources = node['Features']
        if 'Arch' in node:
            currentNode.CpuArch = node['Arch']
        if 'OS' in node:
            currentNode.OSName = node['OS']

        if int(node['TmpDisk']):
            currentNode.TotalFilesystem = float(node['TmpDisk']) * 2**20

        node['State'] = node['State'].lstrip('~')

        if node['State'] in ['IDLE', 'ALLOCATED', 'ALLOCATED+', 'COMPLETING', 'MIXED', 'MIXED+COMPLETING', 'IDLE+POWER', 'IDLE+COMPLETING' ]:
            currentNode.State = 'up'
        elif node['State'] in ['DOWN', 'FAIL', 'FAILING'] or \
                 node['State'][-1] == '*' or \
                 node['State'].find("DOWN") != -1:
            currentNode.State = 'down'
        elif node['State'] in ['DRAINED', 'DRAINING', 'MAINT', 'UNKNOWN', 'IDLE+DRAIN', 'IDLE+DRAIN+POWER'] or \
                 node['State'][-1] == '#' or \
                 node['State'].find("DRAIN") != -1:
            currentNode.State = 'offline'
        else:
                error("Unknown node state: " + node['State'] + " of node: " + node['NodeName'])
                currentNode.State = 'offline'
        
        if currentNode.HostName in jobs:
            currentNode.Jobs = jobs[currentNode.HostName]
        if currentNode.HostName in partition:
            currentNode.Queues = partition[currentNode.HostName]

        __hostsInfoList.append(currentNode)
        __hostsInfoDict[currentNode.HostName] = currentNode

    return(__hostsInfoList,__hostsInfoDict)

def __decodeNodeJobs(output):
    jobs = {}

    for jobInfo in output.splitlines():
        jobInfo = jobInfo.split()

        jobId = jobInfo[0]
        jobNodes = __decodeHostnames(jobInfo[1])

        for node in jobNodes:
            if not node in jobs:
                jobs[node] = []

            jobs[node].append(jobId)

    return jobs

def __decodePartitionNodes(output):
    partition = {}

    partDescList = __formatOutputDict(output)

    for p in partDescList:
        for node in __decodeHostnames(p['Nodes']):
            if not node in partition:
                partition[node] = []

            partition[node].append(p['PartitionName'])

    return partition

def getJobData(jobids=None):
    global __jobsInfoList 
    global __jobsInfoDict 
     
    if __jobsInfoList is not None: 
            return(__jobsInfoList,__jobsInfoDict) 

    showJobOutput = _runCommand('scontrol', ['-a', 'show', 'job', '-o'])
    if re.search("No jobs in the system", showJobOutput):
        return(dict(),[])

    jobDescList = __formatOutputDict(showJobOutput)

    __jobsInfoDict = dict() 
    __jobsInfoList = [] 

    for job in jobDescList:
        currentJob = JobInfo()
        currentJob.JobID = job['JobId']
        currentJob.JobName = job['Name']
        currentJob.OwnerName = job['UserId'][0:job['UserId'].find('(')]
        currentJob.QueueName = job['Partition']
        currentJob.SubmissionTime = __decodeDatetime(job['SubmitTime'])
        if job['JobState'] in ['PENDING', 'CONFIGURING']:
            currentJob.JobStatus = 'Pending'
        elif job['JobState'] in ['RUNNING', 'COMPLETING']:
            currentJob.JobStatus = 'Running'
        elif job['JobState'] in ['CANCELLED', 'COMPLETED', 'FAILED', 'NODE_FAIL', 'TIMEOUT']:
            currentJob.JobStatus = 'Done'
        elif job['JobState'] == 'SUSPENDED':
            currentJob.JobStatus = 'Suspended'
        else:
            currentJob.JobStatus = 'Unknown'

        if job['StartTime'] != 'Unknown':
            if currentJob.JobStatus == "Pending":
                currentJob.EstStartTime = __decodeDatetime(job['StartTime'])
            else:
                currentJob.StartTime = __decodeDatetime(job['StartTime'])

        if job['EndTime'] != 'Unknown' and currentJob.JobStatus == "Done":
            currentJob.EndTime = __decodeDatetime(job['EndTime'])


        # problem with obtaining an information about used slots
        if job['NodeList'] is not None: 
            for host in __decodeHostnames(job['NodeList']):
                currentJob.ExecutionHosts[host] = None

        if job['Reservation'] is not None:
            currentJob.ReservationID = job['Reservation']

        __jobsInfoList.append(currentJob)
        __jobsInfoDict[currentJob.JobID] = currentJob 

    return(__jobsInfoList,__jobsInfoDict)

def addExtendedJobData(job_ids, all_job_ids):
    pass #TODO

def getQueueData(queuenames=None):
    global __queuesInfoList 
    global __queuesInfoDict 
    if __queuesInfoList is not None: 
         return(__queuesInfoList, __queuesInfoDict) 

    showPartitionOutput = _runCommand('scontrol', ['-a', 'show', 'partition', '-o'])
    partDescList = __formatOutputDict(showPartitionOutput)

    jobListOutput = _runCommand('squeue', ['-a', '--state=all', '-h', '--format=%i %P %t'])
    jobs = __decodeJobList(jobListOutput)

    __queuesInfoDict = dict()    
    __queuesInfoList = [] 

    for p in partDescList:
        currentPart = QueueInfo()
        currentPart.QueueName = p['PartitionName']
        currentPart.QueuePriority = int(p['Priority'])
        if p['State'] == 'UP':
            currentPart.Started = True
            currentPart.Enabled = True
        elif p['State'] == 'DOWN':
            currentPart.Started = False
            currentPart.Enabled = True
        elif p['State'] == 'DRAIN':
            currentPart.Started = True
            currentPart.Enabled = False
        else:
            currentPart.Started = False
            currentPart.Enabled = False
 
        currentPart.IsDefault = bool(p['Default'] == 'YES')
        if p['MaxTime'] != 'UNLIMITED':
            currentPart.Walltime = __decodeTimeDelta(p['MaxTime'])
        else:
            currentPart.Walltime = datetime.timedelta(0)

        if currentPart.QueueName in jobs:
            currentPart.PendingJobs = len(jobs[currentPart.QueueName][0])
            currentPart.RunningJobs = len(jobs[currentPart.QueueName][1])

            currentPart.JobIDs = jobs[currentPart.QueueName][0] + jobs[currentPart.QueueName][1]

        __queuesInfoList.append(currentPart)
        __queuesInfoDict[currentPart.QueueName] = currentPart

    return(__queuesInfoList, __queuesInfoDict)

def __decodeTimeDelta(time):
    time = time.split('-')
    days = 0
    if len(time) > 1:
        days = int(time[0])
        time = [time[1]]
    else:
        days = 0

    hours, minutes, seconds = time[0].split(':')

    return datetime.timedelta(days=int(days), hours=int(hours), minutes=int(minutes), seconds=int(seconds))

def __decodeJobList(output):
    jobs = {}

    for job in output.splitlines():
        jobId, part, status = job.split()

        if not part in jobs:
            jobs[part] = [[], []]

        if status in ['CF', 'PD', 'S']:    # pending
            jobs[part][0].append(jobId)
        elif    status in ['R', 'CG']:         # running
            jobs[part][1].append(jobId)

    return jobs

def getReservationData(reservation_names=None):
    global __reservationsInfoList 
    global __reservationsInfoDict 

    if __reservationsInfoList is not None: 
            return(__reservationsInfoList, __reservationsInfoDict) 

    __reservationsInfoList = [] 
    __reservationsInfoDict = dict() 

    showResOutput = _runCommand('scontrol', ['-a', 'show', 'reservation', '-o'])
    if re.search("No reservations in the system", showResOutput):
        return ([],dict())

    resDescList = __formatOutputDict(showResOutput)

    showNodeOutput = _runCommand('scontrol', ['-a', 'show', 'node', '-o']) #TODO cache
    nodeDescList = __formatOutputDict(showNodeOutput)
    nodeDescDict = dict([(n['NodeName'], n) for n in nodeDescList])


    for res in resDescList:
        currentRes = ReservationInfo()
        currentRes.ReservationName = res['ReservationName']
        currentRes.StartTime = __decodeDatetime(res['StartTime'])
        currentRes.EndTime = __decodeDatetime(res['EndTime'])
        if 'MAINT' in res['Flags'].split(','):
            currentRes.Type = 'sr'
        else:
            currentRes.Type = 'ar'

        if res['Users'] is not None:
            currentRes.Owner = res['Users'].split(',')[0]
        else:
            currentRes.Owner = res['Accounts'].split(',')[0]

        if res['Nodes'] is not None:
                for node in __decodeHostnames(res['Nodes']):
                        if node in nodeDescDict and 'CPUTot' in nodeDescDict[node]:
                                currentRes.Resources[node] = int(nodeDescDict[node]['CPUTot'])

        __reservationsInfoList.append(currentRes)
        __reservationsInfoDict[currentRes.ReservationName] = currentRes

    return(__reservationsInfoList, __reservationsInfoDict)
