# $Id: reservation_sge.py 2588 2009-07-28 11:03:09Z mmamonski $
#
# vim: set ts=4 sw=4 et:

#
# QosCosGrid Computing
# Copyright (C) 2008-2011 Poznan Supercomputing and Networking Center
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>
#

try:
	from lxml import etree
except ImportError:
	try:
		# Python 2.5
		import xml.etree.cElementTree as etree
	except ImportError:
		try:
			# Python 2.5
			import xml.etree.ElementTree as etree
		except ImportError:
			try:
				# normal cElementTree install
				import cElementTree as etree
			except ImportError:
				try:
					# normal ElementTree install
					import elementtree.ElementTree as etree
				except ImportError:
					error("Failed to import ElementTree from any known place")
					raise

import os, re, math, string, datetime, time
from commands import getstatusoutput
try:
	from qcg.core import *
	from qcg.comp import *
except ImportError:
	class Module:
		pass

#ns = "{" + SMC_NAMESPACE_ARDL + "}"
ns = "{http://schemas.qoscosgrid.org/comp/2011/04/ardl}"

class InternalFaultException(Exception):
	pass

class NotAuthorizedException(Exception):
	pass

class UnsupportedFeatureException(Exception):
	pass

class InvalidRequestMessageException(Exception):
	pass

class UnknownReservationIdentifierException(Exception):
	pass

class NoSlotsAvailableException(Exception):
	pass

class NoMatchinResourcesException(Exception):
	pass

class ReservationLoadl(Module):

	xmlTimeFormat = "%Y-%m-%dT%H:%M:%S"
	_nodeInfoDict = None

	def formatTime (self, dt):
		from calendar import timegm
		from time import strptime, strftime, localtime

		# The XML dateTime value is always in UTC and with no fractions because it is
		# serialized and then deserialized by gSOAP which uses this constant format.
		utc_tuple = strptime(dt[0:19], self.xmlTimeFormat)
		utc_seconds = timegm(utc_tuple)
		local_tuple = localtime(utc_seconds)
		return strftime("%m/%d/%Y %H:%M", local_tuple)


	# create new reservation
	def create(self, ardl):
		try:
			res = self.internal_create(ardl)
			info ("resId = %s" % (res))
			return (RetCode(RESERVATION_OK), res)
		except NotAuthorizedException, e:
			return RetCode(RESERVATION_NOT_AUTHORIZED_FAULT, str(e) or repr(e))
		except InvalidRequestMessageException, e:
			return RetCode(RESERVATION_INVALID_REQUEST_MESSAGE_FAULT, str(e) or repr(e))
		except NoSlotsAvailableException:
			return RetCode(RESERVATION_NO_SLOTS_AVAILABLE_FAULT, str(e) or repr(e))
		except NoMatchinResourcesException, e:
			return RetCode(RESERVATION_NO_MATCHING_RESOURCES_FAULT, str(e) or repr(e))
		except UnsupportedFeatureException, e:
			return RetCode(RESERVATION_UNSUPPORTED_FEATURE_FAULT, str(e) or repr(e))
		except Exception, e:
			return RetCode(RESERVATION_INTERNAL_FAULT, str(e) or repr(e))


	# terminate reservation
	def terminate(self, arid):
		try:
			self.internal_terminate(arid)
			return RetCode(RESERVATION_OK)
		except InvalidRequestMessageException, e:
			return RetCode(RESERVATION_INVALID_REQUEST_MESSAGE_FAULT, str(e) or repr(e))
		except UnknownReservationIdentifierException:
			return RetCode(RESERVATION_UNKNOWN_RESERVATION_IDENTIFIER_FAULT, arid)
		except NotAuthorizedException, e:
			return RetCode(RESERVATION_NOT_AUTHORIZED_FAULT, str(e) or repr(e))
		except Exception, e:
			return RetCode(RESERVATION_INTERNAL_FAULT, str(e) or repr(e))


	# get list of all reservations
	def get_active(self):
		try:
			res = self.internal_get_active()
			return [RetCode(RESERVATION_OK)] + res
		except Exception, e:
			return RetCode(RESERVATION_INTERNAL_FAULT, str(e) or repr(e))


	# get list of nodes in reservation with cpus total:allocated:free
	def get_status(self, arid):
		try:
			res = self.internal_get_status(arid)
			return (RetCode(RESERVATION_OK), res)
		except InvalidRequestMessageException, e:
			return RetCode(RESERVATION_INVALID_REQUEST_MESSAGE_FAULT, str(e) or repr(e))
		except UnknownReservationIdentifierException:
			return RetCode(RESERVATION_UNKNOWN_RESERVATION_IDENTIFIER_FAULT, arid)
		except Exception, e:
			return RetCode(RESERVATION_INTERNAL_FAULT, str(e) or repr(e))


	def internal_create(self, ardl):
		parsed = etree.fromstring(ardl)
		descr = parsed.find(ns + "ReservationDescription")
		indentification = descr.find(ns + "ReservationIdentification")
		resources = descr.find(ns + "Resources")

		# TODO
		# Project ReservationName
		unsupported = ("CPUArchitecture", "FileSystem", "NativeResource", "OperatingSystem")

		for u in unsupported:
			if resources.find(ns + u) is not None:
				raise UnsupportedFeatureException(u)

		try:
			if resources.find(ns + "ExclusiveExecution").text == "true" or \
			   resources.find(ns + "ExclusiveExecution").text == "1":
				raise UnsupportedFeatureException ("ExclusiveExecution")
		except UnsupportedFeatureException, e:
			raise e
		except:
			pass

		cmd = "llmkres"

		# users ---------------------------------
		try:
			user = resources.find(ns + "UserName").text
			cmd += "  -U " + user
		except:
			pass

		# group ---------------------------------
		try:
			group = resources.find(ns + "GroupName").text
			cmd += "  -G " + group
		except:
			pass

		# time window ---------------------------
		try:
			timeWindow = descr.find(ns + "TimeWindow")

			openWindow = timeWindow.get("open")
			if openWindow is not None and openWindow == "true":
				raise UnsupportedFeatureException("Open TimeWindow")

			startTimeText = timeWindow.find(ns + "StartTime").text
			endTimeText = timeWindow.find(ns + "EndTime").text

			startTime = datetime.datetime (*(time.strptime (startTimeText[:19], self.xmlTimeFormat)[0:6]))
			endTime = datetime.datetime (*(time.strptime (endTimeText[:19], self.xmlTimeFormat)[0:6]))

			startTimeFormated = self.formatTime (startTimeText)

			duration = endTime - startTime
			duration = duration.days * 1440 + duration.seconds / 60

			cmd += "  -t " + startTimeFormated + "  -d " + str (duration)

		except UnsupportedFeatureException, e:
			raise e
		except Exception, e:
			raise InvalidRequestMessageException ("Invalid StartTime or EndTime")

		# slots ---------------------------------
		slotsCount = 0
		try:
			slotsCount = int(resources.find(ns + "ReservedSlotsCount").text)
		except:
			pass

		# additional slots ----------------------
		additionalSlotsCount = 0
		try:
			additionalSlotsCount = int(resources.find(ns + "AdditionalReservedSlotsCount").text)
		except:
			pass

		# memory --------------------------------
		physMemory = 0
		try:
			physMemory = int(resources.find(ns + "IndividualPhysicalMemory").text)
		except:
			pass

		# swap --------------------------------
		swapMemory = 0
		try:
			swapMemory = int(resources.find(ns + "IndividualSwapMemory").text)
		except:
			pass

		# hosts ---------------------------------
		hosts = []
		try:
			candidateHosts = resources.find(ns + "CandidateHosts")

			hosts = [host.text for host in candidateHosts.findall(ns + "HostName")]
		except:
			pass

		output = ""
		if len(hosts):
			command = cmd + "  -h " + " ".join (hosts)
			resId, output = self._verify_reservation (command, slotsCount, physMemory, swapMemory)

			if resId is not None:
				return resId
		else:
			nodeInfoDict = self._get_node_info ()

			# minimum number of cpus on all nodes
			cpusPerNode = min([prop[0] for prop in nodeInfoDict.itervalues()])

			minimalNodesCount = int (math.ceil (float (slotsCount) / cpusPerNode))

			# allowed node count are multiple of cpusPerNode (less than total node count)
			# in range [slotsCount, slotsCount + additionalSlotsCount]
			nodeCountList = xrange (int (min (math.floor ((slotsCount + additionalSlotsCount) / cpusPerNode),
			                                  len (nodeInfoDict))),
			                        minimalNodesCount - 1, -1)

			debug ("cpusPerNode = %d, minimalNodesCount = %d" % (cpusPerNode, minimalNodesCount))
			debug ("nodeCountList = %s" % (nodeCountList))

			# if there is no multiple of cpusPerNode in range
			if len (nodeCountList) == 0:
				# if minimal node count is less than total
				if minimalNodesCount <= len (nodeInfoDict):
					nodeCountList = [minimalNodesCount]
				else:
					debug ("minimalNodesCount > totalNodesCount")
					raise NoMatchinResourcesException ("Not enough nodes")

			for nodeCount in nodeCountList:
				command = cmd + "  -n " + str (nodeCount)
				resId, output = self._verify_reservation (command, slotsCount, physMemory, swapMemory)

				if resId is not None:
					return resId

		if re.search ("RESERVATION_NO_MACHINE", output):
			raise InvalidRequestMessageException ("Invalid HostName")
		if re.search ("RESERVATION_TOO_CLOSE", output):
			raise InvalidRequestMessageException ("The request violates time restrictions")
		if re.search ("RESERVATION_REQUEST_DATA_NOT_VALID", output):
			match = re.search(r'(?P<cmd>\w+): (?P<code>\d+\-\d+) (?P<error>.+)', output)
			if match is not None:
				raise InvalidRequestMessageException (match.group('error'))
			else:
				raise InvalidRequestMessageException

		if re.search ("RESERVATION_NO_RESOURCE", output):
			raise NoMatchinResourcesException ("Insufficient resources to meet the request or resources are busy")

		if re.search ("RESERVATION_NO_PERMISSION", output):
			raise NotAuthorizedException
		if re.search ("RESERVATION_LIMIT_EXCEEDED", output):
			raise NotAuthorizedException ("Request exceeds the maximum number of reservations allowed for the cluster")
		if re.search ("RESERVATION_USER_LIMIT_EXCEEDED", output):
			raise NotAuthorizedException ("Request exceeds the maximum number of reservations allowed for the user")
		if re.search ("RESERVATION_GROUP_LIMIT_EXCEEDED", output):
			raise NotAuthorizedException ("Request exceeds the maximum number of reservations allowed for the group")

		match = re.search(r'(?P<cmd>\w+): (?P<code>\d+\-\d+) (?P<error>.+)', output)
		if match is not None:
			raise InternalFaultException (match.group('error'))
		else:
			raise InternalFaultException


	def internal_get_active (self):
		output = getstatusoutput("llqres -r")

		if output[0] != 0:
			debug ("llqres output code = %d" % (output[0]))
			raise InternalFaultException (output[1])

		if re.search ("There is currently no reservation status to report", output[1]):
			return []

		lines = output[1].splitlines()

		return [line.split('!')[0] for line in lines]


	def internal_get_status (self, arid):
		output = getstatusoutput ("llqres -r -R %s" % (arid))

		if output[0] != 0:
			debug ("llqres output code = %d" % (output[0]))
			if re.search ('errors while processing the list of reservation IDs', output[1]):
				raise InvalidRequestMessageException ("Invalid reservation ID")

			raise InternalFaultException (output[1])

		if re.search ("There is currently no reservation status to report", output[1]):
			raise UnknownReservationIdentifierException

		nodeInfoDict = self._get_node_info()

		resNodes = output[1].split('!')[17].split(',')

		resNodesCpus = []
		for node in resNodes:
			resNodesCpus.append ("%s:%d:%d:0" % (node, nodeInfoDict[node][0], nodeInfoDict[node][0]))

		return resNodesCpus


	def internal_terminate (self, arid):
		output = getstatusoutput ("llrmres -R %s" % (arid))

		if output[0] != 0:
			debug ("llrmres output code = %d" % (output[0]))
			if re.search ('ID .* must be in the format', output[1]):
				raise InvalidRequestMessageException ("Invalid reservation ID")

			raise InternalFaultException (output[1])


	# return dict where keys are node name and value is tuple
	# (cpus, phys memory [MB], virt memory [kB])
	def _get_node_info (self):
		if self._nodeInfoDict is None:
			# -r properties separated by an exclamation point (!)
			# "%n %cpu %m %v" % (machine name, number of cpus, physical memory, available swap space)
			llstatusOut = getstatusoutput ("llstatus -r %n %cpu %m %v")

			if llstatusOut[0] != 0:
				debug ("llstatus output code = %d" % (output[0]))
				raise InternalFaultException (llstatusOut[1])
			llstatusOut = llstatusOut[1].splitlines()

			self._nodeInfoDict = {}
			for nodeInfoStr in llstatusOut:
				nodeInfo = nodeInfoStr.split('!')
				self._nodeInfoDict[nodeInfo[0]] = [int(prop) for prop in nodeInfo[1:]]

		return self._nodeInfoDict


	# execute reservation creation command and verify reserved resources
	# return tuple (resId or None, command output)
	def _verify_reservation (self, command, slotsCount, physMemory, swapMemory):
		info ("cmd = " + command)
		output = getstatusoutput (command)

		match = re.search(r'The reservation (?P<res_id>(\w|\.)+) has been successfully made.', output[1])
		if match is not None:
			# checking if nodes resources are sufficient for requested reservation
			output = getstatusoutput ("llqres -r -R %s" % (match.group('res_id')))

			if output[0] != 0:
				debug ("llqres output code = %d" % (output[0]))
				raise InternalFaultException (output[1])

			resNodes = output[1].split('!')[17].split(',')

			sufficient = True
			cpuSum = 0
			nodeInfoDict = self._get_node_info ()
			for node in resNodes:
				cpuSum += nodeInfoDict[node][0]
				if physMemory > nodeInfoDict[node][1] * 2**20 or \
				   swapMemory > nodeInfoDict[node][2] * 2**10:
					sufficient = False
					break

			if sufficient and slotsCount <= cpuSum:
				return match.group('res_id'), output[1]
			else:
				self.terminate (match.group('res_id'))
				raise NoMatchinResourcesException ("Insufficient resources to meet the request or resources are busy")

		return None, output[1]


if __name__ == '__main__':
	_res = ReservationLoadl()

	print "Creating...",
	_resId = _res.create(open('oneslot.xml', 'r').read())
	if _resId is not None:
		print _resId[0]

	#print "Active:", _res.get_active()

	print "Status:",
	for active in _res.get_active():
		print active, _res.get_status (active)

	if _resId is not None:
		print "Terminating...", _res.terminate(_resId[1])
