# $Id: reservation_sge.py 2588 2009-07-28 11:03:09Z mmamonski $
#
# vim: set ts=4 sw=4 et:

#
# QosCosGrid Computing
# Copyright (C) 2008-2011 Poznan Supercomputing and Networking Center
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>
#

try:
    from lxml import etree
except ImportError:
    try:
        # Python 2.5
        import xml.etree.cElementTree as etree
    except ImportError:
        try:
            # Python 2.5
            import xml.etree.ElementTree as etree
        except ImportError:
            try:
                # normal cElementTree install
                import cElementTree as etree
            except ImportError:
                try:
                    # normal ElementTree install
                    import elementtree.ElementTree as etree
                except ImportError:
                    error("Failed to import ElementTree from any known place")
                    raise

import string
from commands import getstatusoutput
import re
try:
    from qcg.core import *
    from qcg.comp import *
except ImportError:
    class Module:
        pass

#ns = "{" + SMC_NAMESPACE_ARDL + "}"
ns = "{http://schemas.qoscosgrid.org/comp/2011/04/ardl}"

class InternalFaultException(Exception):
    pass

class NotAuthorizedException(Exception):
    pass

class UnsupportedFeatureException(Exception):
    pass

class InvalidRequestMessageException(Exception):
    pass

class UnknownReservationIdentifierException(Exception):
    pass

class NoSlotsAvailableException(Exception):
    pass

class NoMatchinResourcesException(Exception):
    pass

class ReservationPBS(Module):

    def convertTime(self, dt):
        from calendar import timegm
        from time import strptime, strftime, localtime

        # The XML dateTime value is always in UTC and with no fractions because it is
        # serialized and then deserialized by gSOAP which uses this constant format.
        utc_tuple = strptime(dt[0:19], "%Y-%m-%dT%H:%M:%S")
        utc_seconds = timegm(utc_tuple)
        local_tuple = localtime(utc_seconds)
        return strftime("%Y%m%d%H%M.%S", local_tuple)

    def create(self, ardl):
        try:
            res = self.internal_create(ardl)
            return (RetCode(RESERVATION_OK), res)
        except NotAuthorizedException:
            return RetCode(RESERVATION_NOT_AUTHORIZED_FAULT, "Not authorized")
        except InvalidRequestMessageException, e:
            return RetCode(RESERVATION_INVALID_REQUEST_MESSAGE_FAULT, str(e) or repr(e))
        except NoSlotsAvailableException:
            return RetCode(RESERVATION_NO_SLOTS_AVAILABLE_FAULT, "No slots available")
        except NoMatchinResourcesException:
            return RetCode(RESERVATION_NO_MATCHING_RESOURCES_FAULT, "No matching resources")
        except UnsupportedFeatureException, e:
            return RetCode(RESERVATION_UNSUPPORTED_FEATURE_FAULT, str(e))
        except Exception, e:
            return RetCode(RESERVATION_INTERNAL_FAULT, str(e) or repr(e))

    def terminate(self, arid):
        try:
            self.internal_terminate(arid)
            return RetCode(RESERVATION_OK)
        except UnknownReservationIdentifierException:
            return RetCode(RESERVATION_UNKNOWN_RESERVATION_IDENTIFIER_FAULT, arid)
        except NotAuthorizedException, e:
            return RetCode(RESERVATION_NOT_AUTHORIZED_FAULT, str(e) or repr(e))
        except Exception, e:
            return RetCode(RESERVATION_INTERNAL_FAULT, str(e) or repr(e))

    def get_active(self):
        try:
            res = self.internal_get_active()
            return [RetCode(RESERVATION_OK)] + res
        except Exception, e:
            return RetCode(RESERVATION_INTERNAL_FAULT, str(e) or repr(e))

    def get_status(self, arid):
        try:
            res = self.internal_get_status(arid)
            return (RetCode(RESERVATION_OK), res)
        except UnknownReservationIdentifierException:
            return RetCode(RESERVATION_UNKNOWN_RESERVATION_IDENTIFIER_FAULT, arid)
        except Exception, e:
            return RetCode(RESERVATION_INTERNAL_FAULT, str(e) or repr(e))

    def internal_create(self, ardl):
        parsed = etree.fromstring(ardl)
        descr = parsed.find(ns + "ReservationDescription")
        resources = descr.find(ns + "Resources")

        # TODO
        unsupported = ("FileSystem", "OperatingSystem", "CPUArchitecture", "NativeResource")

        for u in unsupported:
            if resources.find(ns + u) is not None:
                raise UnsupportedFeatureException(u)

        if resources.find(ns + "ExclusiveExecution") is not None:
            if resources.find(ns + "ExclusiveExecution").text == "true":
                raise UnsupportedFeatureException("ExclusiveExecution")

        cmd = "pbs_rsub -m n -I -30"

        try:
            #The name specified may be up to 15 characters in length. It must consist of printable,
            #non-white space characters with the first character alphabetic.

            indentification = descr.find(ns + "ReservationIdentification")
            res_name = indentification.find(ns + "ReservationName").text
            res_name = '_'.join(res_name.split())
            res_name = res_name.lstrip(string.digits + string.punctuation)
            res_name = res_name[0:15]

            if len (res_name) > 0:
                cmd += " -N " + res_name
        except:
            pass

        try:
            timeWindow = descr.find(ns + "TimeWindow")

            openWindow = timeWindow.get("open")
            if openWindow is not None and openWindow == "true":
                raise UnsupportedFeatureException("Open TimeWindow")

            startTime = timeWindow.find(ns + "StartTime").text
            endTime = timeWindow.find(ns + "EndTime").text

            startTime = self.convertTime(startTime)
            endTime = self.convertTime(endTime)

            cmd += " -R " + startTime + " -E " + endTime
        except UnsupportedFeatureException, e:
            raise e
        except Exception, e:
            raise InvalidRequestMessageException("No StartTime or EndTime: " + str(e) or repr(e))

        try:
            cmd += " -U+" + resources.find(ns + "UserName").text
        except:
            pass

        try:
            cmd += " -G+" + resources.find(ns + "GroupName").text
        except:
            pass

        try:
            cmd += " -l pmem=" + resources.find(ns + "IndividualPhysicalMemory").text
        except:
            pass

        try:
            cmd += " -l pvmem=" + resources.find(ns + "IndividualSwapMemory").text
        except:
            pass

        try:
            candidateHosts = resources.find(ns + "CandidateHosts")

            if candidateHosts:
                firstRun = True

                for hostName in candidateHosts.findall(ns + "HostName"):
                    if not firstRun:
                        raise UnsupportedFeatureException("Multiple HostName selection is currently unsupported")
                    else:
                        firstRun = False

                #cmd += ":host=" + hostName.text
        except:
            pass

        slotsCount = 1
        try:
            slotsCount = resources.find(ns + "ReservedSlotsCount").text
        except:
            pass

        nodesCount = None
        try:
            nodesCount = resources.find(ns + "ReservedNodesCount").text
        except:
            pass

        slotsPerNode = 1
        try:
            slotsPerNode = resources.find(ns + "ReservedSlotsPerNode").text
        except:
            pass


        additionalSlotsCount = 0
        try:
            additionalSlotsCount = resources.find(ns + "AdditionalReservedSlotsCount").text
        except:
            pass

        if nodesCount is not None:
            command = cmd + " -l select=" + str(nodesCount) + ":ncpus=" + str(slotsPerNode)
            info("cmd = " + command)
            output = getstatusoutput(command)

            # exit code means nothing....
            if re.search(' CONFIRMED ', output[1]):
                debug("CONFIRMED ")
                return output[1].split('.')[0]
        else:
            for slots in range((int(additionalSlotsCount) + int(slotsCount)), (int(slotsCount) - 1), -1):
                command = cmd + " -l select=" + str(slots) + ":ncpus=1"
                info("cmd = " + command)
                output = getstatusoutput(command)

                # exit code means nothing....
                if re.search(' CONFIRMED ', output[1]):
                    debug("CONFIRMED ")
                    #debug("out = " + output[1])
                    return output[1].split('.')[0]

        if re.search("illegal -N value", output[1]):
            raise InvalidRequestMessageException("Illegal reservation name")
        if re.search("Bad time specification", output[1]):
            raise InvalidRequestMessageException("Bad time specification")

        if re.search("DENIED", output[1]): # probably not enough resources
            raise NoSlotsAvailableException("Reservation request DENIED")

        if re.search("wait period expired", output[1]): # probably not enough resources
            raise NoSlotsAvailableException("wait period expired")


        notice("pbs_rsub failed: " + output[1])

        raise InternalFaultException(output[1])

    def internal_get_data(self, command):
        output = getstatusoutput(command)

        if output[0] != 0:
            raise InternalFaultException(output[1])

        return output[1].splitlines()

    def internal_get_active(self):
        lines = self.internal_get_data("pbs_rstat -B")

        return [line.split()[1].split('.')[0] for line in lines]

    def internal_get_status(self, arid):
        lines = self.internal_get_data("pbs_rstat -f " + arid);

        hostsformatted = []
        for line in lines:
            if line.startswith("resv_nodes = "):
                resv_nodes = line[13:].split('+')

        resv_nodes = [n[1:-1] for n in resv_nodes]

        for node in resv_nodes:
            resources = node.split(':')

            host = resources[0].split('.')[0]
            host_info = self.internal_get_data("pbsnodes " + host)
            procsCount = host_info[9].split()[-1]

            for res in resources:
                if res.find('ncpus') != -1:
                    reservedProcs = res[6:]
                    hostsformatted.append (host + ":" + procsCount + ":" +  reservedProcs + ":0")

        assert len(hostsformatted) != 0

        return hostsformatted

    def internal_terminate(self, arid):
        res = getstatusoutput("pbs_rdel " + arid)

        if res[0] == 0:
            return

        if re.search("Unknown Reservation Id", res[1]):
            raise UnknownReservationIdentifierException

        if re.search("Unauthorized Request". res[1]):
            raise NotAuthorizedException

        notice("releaseres failed: " + res[1])

        raise InternalFaultException(res[1])


if __name__ == '__main__':
    _res = ReservationPBS()
    f = open('oneslot.xml', 'r')
    xml_string = f.read()

    print "Creating..."
    _resId = _res.create(xml_string)

    print "Active:"
    _res.get_active()

    print "Status:"
    _res.get_status(_resId)

    print "Terminating..."
    _res.terminate(_resId)
