# $Id: reservation_sge.py 2588 2009-07-28 11:03:09Z mmamonski $
#
# vim: set ts=4 sw=4 et:

#
# QosCosGrid Computing
# Copyright (C) 2008-2011 Poznan Supercomputing and Networking Center
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>
#

try:
	from lxml import etree
except ImportError:
	try:
		# Python 2.5
		import xml.etree.cElementTree as etree
	except ImportError:
		try:
			# Python 2.5
			import xml.etree.ElementTree as etree
		except ImportError:
			try:
				# normal cElementTree install
				import cElementTree as etree
			except ImportError:
				try:
					# normal ElementTree install
					import elementtree.ElementTree as etree
				except ImportError:
					error("Failed to import ElementTree from any known place")
					raise

import os, re, math, string
from commands import getstatusoutput, getoutput
try:
	from qcg.core import *
	from qcg.comp import *
except ImportError:
	class Module:
		pass

	def debug(msg):
		print msg

	def RetCode(code, msg = ""):
		print "RetCode: (%s, %s)" % (code, msg)
		return (code, )

	RESERVATION_OK, \
	RESERVATION_NOT_AUTHORIZED_FAULT, \
	RESERVATION_INVALID_REQUEST_MESSAGE_FAULT, \
	RESERVATION_NO_SLOTS_AVAILABLE_FAULT, \
	RESERVATION_NO_MATCHING_RESOURCES_FAULT, \
	RESERVATION_UNSUPPORTED_FEATURE_FAULT, \
	RESERVATION_INTERNAL_FAULT, \
	RESERVATION_UNKNOWN_RESERVATION_IDENTIFIER_FAULT = range(8)

#ns = "{" + SMC_NAMESPACE_ARDL + "}"
ns = "{http://schemas.qoscosgrid.org/comp/2011/04/ardl}"

class InternalFaultException(Exception):
	pass

class NotAuthorizedException(Exception):
	pass

class UnsupportedFeatureException(Exception):
	pass

class InvalidRequestMessageException(Exception):
	pass

class UnknownReservationIdentifierException(Exception):
	pass

class NoSlotsAvailableException(Exception):
	pass

class NoMatchinResourcesException(Exception):
	pass

class ReservationSlurm(Module):

	regexpNamePattern = "^[a-zA-Z_][a-zA-Z_0-9\-]*$"
	regexpNumberPattern = "^[0-9]+$"

	def convertTime(self, dt):
		from calendar import timegm
		from time import strptime, strftime, localtime

		# The XML dateTime value is always in UTC and with no fractions because it is
		# serialized and then deserialized by gSOAP which uses this constant format.
		utc_tuple = strptime(dt[0:19], "%Y-%m-%dT%H:%M:%S")
		utc_seconds = timegm(utc_tuple)
		local_tuple = localtime(utc_seconds)
		return strftime("%Y-%m-%dT%H:%M:%S", local_tuple)

	def create(self, ardl):
		try:
			res = self.internal_create(ardl)
			return (RetCode(RESERVATION_OK), res)
		except NotAuthorizedException:
			return RetCode(RESERVATION_NOT_AUTHORIZED_FAULT, "Invalid user id")
		except InvalidRequestMessageException, e:
			return RetCode(RESERVATION_INVALID_REQUEST_MESSAGE_FAULT, str(e) or repr(e))
		except NoSlotsAvailableException:
			return RetCode(RESERVATION_NO_SLOTS_AVAILABLE_FAULT, "No slots available")
		except NoMatchinResourcesException, e:
			return RetCode(RESERVATION_NO_MATCHING_RESOURCES_FAULT, str(e) or repr(e))
		except UnsupportedFeatureException, e:
			return RetCode(RESERVATION_UNSUPPORTED_FEATURE_FAULT, str(e))
		except Exception, e:
			return RetCode(RESERVATION_INTERNAL_FAULT, str(e) or repr(e))

	def terminate(self, arid):
		try:
			self.internal_terminate(arid)
			return RetCode(RESERVATION_OK)
		except InvalidRequestMessageException, e:
			return RetCode(RESERVATION_INVALID_REQUEST_MESSAGE_FAULT, str(e) or repr(e))
		except UnknownReservationIdentifierException:
			return RetCode(RESERVATION_UNKNOWN_RESERVATION_IDENTIFIER_FAULT, arid)
		except NotAuthorizedException, e:
			return RetCode(RESERVATION_NOT_AUTHORIZED_FAULT, str(e) or repr(e))
		except Exception, e:
			return RetCode(RESERVATION_INTERNAL_FAULT, str(e) or repr(e))

	def get_active(self):
		try:
			res = self.internal_get_active()
			return [RetCode(RESERVATION_OK)] + res
		except Exception, e:
			return RetCode(RESERVATION_INTERNAL_FAULT, str(e) or repr(e))

	def get_status(self, arid):
		try:
			res = self.internal_get_status(arid)
			return (RetCode(RESERVATION_OK), res)
		except InvalidRequestMessageException, e:
			return RetCode(RESERVATION_INVALID_REQUEST_MESSAGE_FAULT, str(e) or repr(e))
		except UnknownReservationIdentifierException:
			return RetCode(RESERVATION_UNKNOWN_RESERVATION_IDENTIFIER_FAULT, arid)
		except Exception, e:
			return RetCode(RESERVATION_INTERNAL_FAULT, str(e) or repr(e))

	def internal_create(self, ardl):
		parsed = etree.fromstring(ardl)
		descr = parsed.find(ns + "ReservationDescription")
		indentification = descr.find(ns + "ReservationIdentification")
		resources = descr.find(ns + "Resources")

		# TODO
		unsupported = ("FileSystem", "OperatingSystem", "CPUArchitecture", "NativeResource",
		               "IndividualSwapMemory", "CandidateHosts", "GroupName")

		for u in unsupported:
			if resources.find(ns + u) is not None:
				raise UnsupportedFeatureException(u)

		if resources.find(ns + "ExclusiveExecution") is not None:
			if resources.find(ns + "ExclusiveExecution").text == "true":
				raise UnsupportedFeatureException("ExclusiveExecution")

		cmd = "scontrol create reservation"

		# reservation name ----------------------
		try:
			resName = indentification.find(ns + "ReservationName").text
		except:
			pass
		else:
			if not re.match(self.regexpNamePattern, resName):
				raise InvalidRequestMessageException ("Invalid ReservationName")

			cmd += " Reservation=" + resName

		# project -------------------------------
		account = None
		try:
			account = indentification.find(ns + "Project").text
		except:
			pass
		else:
			if not re.match(self.regexpNamePattern, account):
				raise InvalidRequestMessageException ("Invalid Project string")

			cmd += " Accounts=" + account

		# users ---------------------------------
		users = None
		try:
			users = resources.find(ns + "UserName").text
		except:
			pass
		else:
			if not re.match(self.regexpNamePattern, users):
				raise InvalidRequestMessageException ("Invalid UserName")

		if users is not None:
			cmd += " Users=" + users
		elif account is None:
			cmd += " Users=" + os.getenv("USER")

		# time window ---------------------------
		try:
			timeWindow = descr.find(ns + "TimeWindow")

			openWindow = timeWindow.get("open")
			if openWindow is not None and openWindow == "true":
				raise UnsupportedFeatureException("Open TimeWindow")

			startTime = timeWindow.find(ns + "StartTime").text
			endTime = timeWindow.find(ns + "EndTime").text

			startTime = self.convertTime(startTime)
			endTime = self.convertTime(endTime)

			cmd += " StartTime=" + startTime + " EndTime=" + endTime
		except UnsupportedFeatureException, e:
			raise e
		except Exception:
			raise InvalidRequestMessageException("Invalid StartTime or EndTime")

		# slots ---------------------------------
		slotsCount = 0
		try:
			slotsCount = int (resources.find(ns + "ReservedSlotsCount").text)
		except:
			pass
		else:
			if slotsCount < 0:
				raise InvalidRequestMessageException("ReservedSlotsCount must be positive")

		# additional slots ----------------------
		additionalSlotsCount = 0
		try:
			additionalSlotsCount = int (resources.find(ns + "AdditionalReservedSlotsCount").text)
		except:
			pass
		else:
			if additionalSlotsCount < 0:
				raise InvalidRequestMessageException("AdditionalReservedSlotsCount must be positive")

		# memory --------------------------------
		physMemory = 0
		try:
			physMemory = int (resources.find(ns + "IndividualPhysicalMemory").text)
		except:
			pass
		else:
			if physMemory < 0:
				raise InvalidRequestMessageException("IndividualPhysicalMemory must be positive")

		# features from env ---------------------
		features = os.getenv("QCG_AR_SLURM_FEATURES")
		if features is not None and len(features) == 0:
			features = None

		# partition -----------------------------
		if os.getenv("QCG_AR_SLURM_PARTITION") is not None and \
		   len(os.getenv("QCG_AR_SLURM_PARTITION")) > 0:
			partitionName = os.getenv("QCG_AR_SLURM_PARTITION")
		else:
			# get default partition -- with '*' at the end of name
			output = getstatusoutput("sinfo --format=\"%100P\" -h | grep \"*\"")

			if output[0]:
				raise InternalFaultException(output[1])

			partitionName = output[1].splitlines()[0].strip()
			partitionName = partitionName[:-1]

		if not re.match(self.regexpNamePattern, partitionName):
			raise InvalidRequestMessageException ("Invalid partition name")

		# format: node name; node count; cpu per node; memory per node; features
		if features is not None:
			cmd += " Features=" + features

			sinfoCmd = "sinfo -h --format=\"%N %D %c %m %f\" | grep " + features
		else:
			cmd += " PartitionName=" + partitionName

			sinfoCmd = "sinfo -h --format=\"%N %D %c %m\" -p " + partitionName

		nodeInfoOut = getstatusoutput (sinfoCmd)
		#debug (nodeInfoOut)

		if nodeInfoOut[0]:
			raise InternalFaultException(nodeInfoOut[1])
		if len(nodeInfoOut[1]) == 0:
			raise InvalidRequestMessageException("Partition does not exist")

		# assume all nodes has the same parameters as first on list
		nodeInfo = nodeInfoOut[1].splitlines()[0].split()

		NODES, COUNT, CPU, MEM, FEATURES = range (5)

		nodeInfo[COUNT] = int(nodeInfo[COUNT])

		if nodeInfo[CPU][-1] == '+':
			nodeInfo[CPU] = int(nodeInfo[CPU][:-1])
		else:
			nodeInfo[CPU] = int(nodeInfo[CPU])

		if nodeInfo[MEM][-1] == '+':
			nodeInfo[MEM] = int(nodeInfo[MEM][-1])
		else:
			nodeInfo[MEM] = int(nodeInfo[MEM])

		if features is not None:
			nodeInfo[FEATURES] = nodeInfo[FEATURES].split(',')

		cpus = nodeInfo[COUNT] * nodeInfo[CPU]
		mem = nodeInfo[COUNT] * nodeInfo[CPU] * 2**20 # megabyte -> byte

		#debug (nodeInfo)

		if slotsCount > cpus:
			raise NoMatchinResourcesException("Not enough slots")
		if physMemory > mem:
			raise NoMatchinResourcesException("Not enough memory")

		minNodeCount = max(math.ceil(float(slotsCount)/nodeInfo[CPU]),
		                   math.ceil(float(physMemory)/(nodeInfo[MEM] * 2**20)))
		maxNodeCount = max(minNodeCount,
		                   math.floor(float(slotsCount + additionalSlotsCount)/nodeInfo[CPU]))

		#debug ("minNodeCount: " + str(minNodeCount) + ", maxNodeCount: " + str(maxNodeCount))

		if minNodeCount > int(nodeInfo[COUNT]):
			raise NoMatchinResourcesException

		output = ""
		for nodeCount in range(int(maxNodeCount), int(minNodeCount - 1), -1):
			command = cmd + " NodeCnt=" + str(nodeCount)
			debug (command)

			output = getstatusoutput (command)
			if re.search('Reservation created', output[1]):
				return output[1].split(': ')[1]

		if re.search("Invalid time specified", output[1]):
			raise InvalidRequestMessageException("Invalid time specified")
		if re.search("Requested nodes are busy", output[1]):
			raise NoSlotsAvailableException
		if re.search("Invalid user id", output[1]):
			raise NotAuthorizedException

		raise InternalFaultException(output[1])

	def internal_get_active(self):
		output = getstatusoutput("scontrol show reservation -o")

		if output[0] != 0:
			raise InternalFaultException(output[1])

		if re.search ("No reservations in the system", output[1]):
			return []

		lines = output[1].splitlines()

		return [line.split()[0].split('=')[1] for line in lines]

	def internal_get_status(self, arid):
		if not re.match(self.regexpNamePattern, arid):
			raise InvalidRequestMessageException ("Invalid reservation name")

		output = getstatusoutput("scontrol -o show reservation=" + arid)

		if output[0]:
			if re.search("Reservation .* not found", output[1]):
				raise UnknownReservationIdentifierException

			raise InternalFaultException (output[1])

		nodes = output[1].splitlines()[0].split()[4].split("=")[1]

		if len(nodes) == 0:
			return []

		nodes = self._decodeHostnames (nodes)

		hostsformatted = []
		for node in nodes:
			output = getoutput("sinfo --format=\"%c\" -hn " + node)

			hostsformatted.append (node + ":" + output + ":" +  output + ":0")

		assert len(hostsformatted) != 0

		return hostsformatted

	def _decodeHostnames (self, hostnames):
		return getoutput ('scontrol show hostnames ' + hostnames).splitlines()

	def internal_terminate(self, arid):
		if not re.match(self.regexpNamePattern, arid):
			raise InvalidRequestMessageException ("Invalid reservation name")

		output = getoutput("scontrol delete Reservation=" + arid)

		if (len(output) == 0):
			return

		if re.search("Requested reservation is invalid", output):
			raise UnknownReservationIdentifierException

		if re.search("Invalid user id", output):
			raise NotAuthorizedException

		raise InternalFaultException(output)

if __name__ == '__main__':
	_res = ReservationSlurm()

	print "Creating...",
	_resId = _res.create(open('oneslot.xml', 'r').read())
	if isinstance(_resId[0], tuple):
		print _resId[1]

	print "Active:", _res.get_active()[1:]

	print "Status:",
	for active in _res.get_active()[1:]:
		print _res.get_status(active)

	if isinstance(_resId[0], tuple):
		print "Terminating...", _res.terminate(_resId[1])
